from __future__ import annotations

import re
from itertools import groupby


def split_cigar(CIGAR: str) -> list[str]:
    cigar = re.split(r"([MIDNSH=X])", CIGAR)
    return [cigar[i] + cigar[j] for i, j in zip(range(0, len(cigar), 2), range(1, len(cigar), 2))]


def calculate_alignment_length(CIGAR: str) -> int:
    return sum(int(c[:-1]) for c in split_cigar(CIGAR) if c[-1] in "MDN=X")


def is_header(s: list[str]) -> bool:
    return s[0].startswith("@")


def is_mapped(s: list[str]) -> bool:
    return not s[0].startswith("@") and s[9] != "*"


###########################################################
# remove_overlapped_reads
###########################################################


def is_overlapping(prev_alignment: list[str], next_alignment: list[str]) -> bool:
    prev_start = int(prev_alignment[3])
    prev_alignment_len = calculate_alignment_length(prev_alignment[5])
    prev_end = prev_start + prev_alignment_len
    next_start = int(next_alignment[3])
    return prev_end >= next_start


def remove_overlapped_reads(sam: list[list[str]]) -> list[list[str]]:
    sam_headers = [s for s in sam if is_header(s)]
    sam_contents = [s for s in sam if is_mapped(s)]

    sam_contents.sort(key=lambda x: (x[0], int(x[3])))
    sam_trimmed = sam_headers.copy()

    for _, group in groupby(sam_contents, key=lambda x: x[0]):
        alignments = list(group)

        if len(alignments) == 1:
            sam_trimmed.append(alignments[0])
            continue

        if not any(is_overlapping(alignments[idx], alignments[idx + 1]) for idx in range(len(alignments) - 1)):
            sam_trimmed += alignments

    return sam_trimmed


###########################################################
# remove_microholomogy
###########################################################


def trim_softclip(CIGAR: str, SEQ: str) -> str:
    cigar_split = split_cigar(CIGAR)
    if cigar_split[0].endswith("S"):
        SEQ = SEQ[int(cigar_split[0][:-1]) :]
    if cigar_split[-1].endswith("S"):
        SEQ = SEQ[: -int(cigar_split[-1][:-1])]
    return SEQ


def check_microhomology(curr_seq_trimmed: str, next_seq_trimmed: str) -> int:
    len_microhomology = 0
    for i in range(1, min(len(curr_seq_trimmed), len(next_seq_trimmed))):
        if curr_seq_trimmed[-i:] == next_seq_trimmed[:i]:
            len_microhomology = i
    return len_microhomology


def count_mutations_in_microhomology(cigar_split: list, len_microhomology: int) -> int:
    """Count the number of mutations within the microhomology based on a CIGAR list."""
    total_num = 0
    count_mutation = 0
    for cigar in cigar_split:
        num, op = int(cigar[:-1]), cigar[-1]
        total_num += num
        # Stop counting when total_num reaches the length of microhomology
        if total_num >= len_microhomology:
            break
        # Increment mutation count for non-match operations
        if op != "M":
            count_mutation += num
    return count_mutation


def trim_cigar_on_microhomology(cigar_split: list, len_microhomology: int) -> tuple(str, int):
    """Trim CIGAR based on the length of the microhomology."""
    cigar_trimmed = cigar_split[::-1]
    total_num = 0
    num_deleletion = 0
    while total_num < len_microhomology:
        cigar = cigar_trimmed.pop()
        num, op = int(cigar[:-1]), cigar[-1]
        # Skip deletion operations
        if op == "D":
            num_deleletion += num
            continue
        total_num += num
        # Only append to the list when total_num is greater or equal to len_microhomology
        if total_num > len_microhomology:
            num = total_num - len_microhomology
            cigar_trimmed.append(f"{num}{op}")
    return "".join(cigar_trimmed[::-1]), num_deleletion


def trim(alignments: list[list[str]]) -> list[list[str]]:
    idx = 0
    while idx < len(alignments) - 1:
        curr_align, next_align = alignments[idx], alignments[idx + 1]
        curr_cigar, next_cigar = curr_align[5], next_align[5]
        curr_seq, next_seq = curr_align[9], next_align[9]
        next_qual = next_align[10]
        # trim softclip
        curr_seq_trimmed = trim_softclip(curr_cigar, curr_seq)
        next_seq_trimmed = trim_softclip(next_cigar, next_seq)
        next_qual_trimmed = trim_softclip(next_cigar, next_qual)
        # Check the length of microhomology
        len_microhomology = check_microhomology(curr_seq_trimmed, next_seq_trimmed)
        # Continue if no microhomology exists
        if len_microhomology == 0:
            idx += 1
            continue
        # Update CIGAR
        curr_cigar_splitted = [c for c in split_cigar(curr_cigar) if not re.search(r"[SH]$", c)]
        next_cigar_splitted = [c for c in split_cigar(next_cigar) if not re.search(r"[SH]$", c)]
        mutations_in_curr = count_mutations_in_microhomology(curr_cigar_splitted[::-1], len_microhomology)
        mutations_in_next = count_mutations_in_microhomology(next_cigar_splitted, len_microhomology)
        # Update START, CIGAR, SEQ, QUAL to the more mismatched alignment within microhomology
        if mutations_in_next >= mutations_in_curr:
            alignments[idx + 1][5], num_del = trim_cigar_on_microhomology(next_cigar_splitted, len_microhomology)
            alignments[idx + 1][3] = str(int(next_align[3]) + len_microhomology + num_del)
            alignments[idx + 1][9] = next_seq_trimmed[len_microhomology:]
            alignments[idx + 1][10] = next_qual_trimmed[len_microhomology:]
        else:
            alignments[idx][5], num_del = trim_cigar_on_microhomology(next_cigar_splitted, len_microhomology)
            alignments[idx][3] = str(int(alignments[idx][3]) + num_del)
            alignments[idx][9] = next_seq_trimmed[:-len_microhomology]
            alignments[idx][10] = next_qual_trimmed[:-len_microhomology:]
        idx += 1


def remove_microhomology(sam: list[list[str]]) -> list[list[str]]:
    sam_headers = [s for s in sam if is_header(s)]
    sam_contents = [s for s in sam if is_mapped(s)]

    sam_contents.sort(key=lambda x: (x[0], int(x[3])))
    sam_trimmed = sam_headers.copy()

    for _, group in groupby(sam_contents, key=lambda x: x[0]):
        alignments = list(group)
        if len(alignments) == 1:
            sam_trimmed.append(alignments[0])
            continue
        trim(alignments)
        sam_trimmed.extend(alignments)
    return sam_trimmed


###########################################################
# revcomp_sam
###########################################################


def revcomp(sequence: str) -> str:
    """Reverse complement a DNA sequence."""
    complement = {"A": "T", "C": "G", "G": "C", "T": "A"}
    return "".join(complement[nt] for nt in sequence[::-1])


def reverse_flag(flag: int) -> int:
    """Reverse the flag of a SAM entry."""
    return flag | 16 if flag & 16 == 0 else flag & ~16


def revcomp_sam(sam_contents: list[list[str]], genome_end: int) -> list[str]:
    """Reverse and complement SAM entries."""
    sam_reversed = []
    for sam_content in sam_contents:
        sam_update = sam_content.copy()
        sam_update[1] = str(reverse_flag(int(sam_content[1])))
        sam_cigar = "".join(split_cigar(sam_content[5])[::-1])
        sam_update[5] = sam_cigar
        sam_start = int(sam_content[3])
        sam_length = calculate_alignment_length(sam_cigar)
        sam_update[3] = str(genome_end - (sam_start + sam_length) + 2)
        sam_update[9] = revcomp(sam_content[9])
        sam_update[10] = sam_content[10][::-1]
        sam_reversed.append(sam_update)
    return sam_reversed



from __future__ import annotations

import pickle
import shutil
import logging

from pathlib import Path
from datetime import datetime
from typing import NamedTuple
from collections import defaultdict

from DAJIN2.utils import io, config
from DAJIN2.core import classification, clustering, consensus, preprocess, report

logger = logging.getLogger(__name__)


def parse_arguments(arguments: dict) -> tuple:
    genome_urls = defaultdict(str)
    if "genome" in arguments:
        genome_urls.update(
            {"genome": arguments["genome"], "blat": arguments["blat"], "goldenpath": arguments["goldenpath"]}
        )

    return (
        arguments["sample"],
        arguments["control"],
        arguments["allele"],
        arguments["name"],
        arguments["threads"],
        genome_urls,
    )


def convert_inputs_to_posix(sample: str, control: str, allele: str) -> tuple:
    sample = io.convert_to_posix(sample)
    control = io.convert_to_posix(control)
    allele = io.convert_to_posix(allele)
    return sample, control, allele


def create_temporal_directory(name: str, control_name: str) -> Path:
    tempdir = Path(config.TEMP_ROOT_DIR, name)
    Path(tempdir, "cache", ".igvjs", control_name).mkdir(parents=True, exist_ok=True)
    return tempdir


def check_caches(tempdir: Path, path_allele: str, genome_url: str) -> bool:
    is_cache_hash = preprocess.cache_checker.exists_cached_hash(tempdir=tempdir, path=path_allele)
    is_cache_genome = preprocess.cache_checker.exists_cached_genome(tempdir=tempdir, genome=genome_url)
    return is_cache_hash and is_cache_genome


def get_genome_coordinates(genome_urls: dict, fasta_alleles: dict, is_cache_genome: bool, tempdir: Path) -> dict:
    genome_coordinates = {
        "genome": genome_urls["genome"],
        "chrom_size": 0,
        "chrom": "control",
        "start": 0,
        "end": len(fasta_alleles["control"]) - 1,
        "strand": "+",
    }
    if genome_urls["genome"]:
        if is_cache_genome:
            genome_coordinates = list(io.read_jsonl(Path(tempdir, "cache", "genome_coodinates.jsonl")))[0]
        else:
            genome_coordinates = preprocess.genome_fetcher.fetch_coordinates(
                genome_coordinates, genome_urls, fasta_alleles["control"]
            )
            genome_coordinates["chrom_size"] = preprocess.genome_fetcher.fetch_chromosome_size(
                genome_coordinates, genome_urls
            )
            io.write_jsonl([genome_coordinates], Path(tempdir, "cache", "genome_coodinates.jsonl"))
    return genome_coordinates


class FormattedInputs(NamedTuple):
    path_sample: str
    path_control: str
    path_allele: str
    sample_name: str
    control_name: str
    fasta_alleles: dict[str, str]
    tempdir: Path
    genome_coordinates: dict[str, str]
    threads: int


def format_inputs(arguments: dict) -> FormattedInputs:
    path_sample, path_control, path_allele, name, threads, genome_urls = parse_arguments(arguments)
    path_sample, path_control, path_allele = convert_inputs_to_posix(path_sample, path_control, path_allele)
    sample_name = preprocess.fastx_parser.extract_basename(path_sample)
    control_name = preprocess.fastx_parser.extract_basename(path_control)
    fasta_alleles = preprocess.fastx_parser.dictionize_allele(path_allele)
    tempdir = create_temporal_directory(name, control_name)
    is_cache_genome = check_caches(tempdir, path_allele, genome_urls["genome"])
    genome_coordinates = get_genome_coordinates(genome_urls, fasta_alleles, is_cache_genome, tempdir)
    return FormattedInputs(
        path_sample,
        path_control,
        path_allele,
        sample_name,
        control_name,
        fasta_alleles,
        tempdir,
        genome_coordinates,
        threads,
    )


def _dtnow() -> str:
    return datetime.now().strftime("%Y-%m-%d %H:%M:%S")


###########################################################
# main
###########################################################


def execute_control(arguments: dict):
    logger.info(f"{_dtnow()}: {arguments['control']} is now processing...")
    ###########################################################
    # Preprocess
    ###########################################################
    ARGS = format_inputs(arguments)
    preprocess.directories.create_temporal(ARGS.tempdir, ARGS.control_name, is_control=True)
    preprocess.directories.create_report(ARGS.tempdir, ARGS.control_name, is_control=True)
    io.cache_control_hash(ARGS.tempdir, ARGS.path_allele)
    ###########################################################
    # Check caches
    ###########################################################
    if Path(ARGS.tempdir, "report", "BAM", ARGS.control_name, f"{ARGS.control_name}.bam").exists():
        logger.info(f"{arguments['control']} is already preprocessed and reuse the results for the current run...")
        return
    logger.info(f"{_dtnow()}: Preprocess {arguments['control']}...")
    ###########################################################
    # Mapping
    ###########################################################
    # ============================================================
    # Export fasta files as single-FASTA format
    # ============================================================
    preprocess.fastx_parser.export_fasta_files(ARGS.tempdir, ARGS.fasta_alleles, ARGS.control_name)
    # ============================================================
    # Mapping using mappy
    # ============================================================
    paths_fasta = Path(ARGS.tempdir, ARGS.control_name, "fasta").glob("*.fasta")
    preprocess.mapping.generate_sam(ARGS.tempdir, paths_fasta, ARGS.path_control, ARGS.control_name, ARGS.threads)
    ###########################################################
    # MIDSV conversion
    ###########################################################
    preprocess.midsv_caller.execute(ARGS.tempdir, ARGS.fasta_alleles, ARGS.control_name)
    ###########################################################
    # Prepare data to `extract mutaion loci`
    ###########################################################
    preprocess.extract_mutation_loci(
        ARGS.tempdir, ARGS.fasta_alleles, ARGS.sample_name, ARGS.control_name, is_control=True
    )
    ###########################################################
    # Output BAM
    ###########################################################
    logger.info(f"{_dtnow()}: Output BAM files of {arguments['control']}...")
    report.report_bam.output_bam(
        ARGS.tempdir, ARGS.control_name, ARGS.genome_coordinates, ARGS.threads, is_control=True
    )
    ###########################################################
    # Finish call
    ###########################################################
    logger.info(f"{_dtnow()}: \N{teacup without handle} {arguments['control']} is finished!")


def execute_sample(arguments: dict):
    logger.info(f"{_dtnow()}: {arguments['sample']} is now processing...")
    ###########################################################
    # Preprocess
    ###########################################################
    ARGS = format_inputs(arguments)
    preprocess.directories.create_temporal(ARGS.tempdir, ARGS.sample_name)
    preprocess.directories.create_report(ARGS.tempdir, ARGS.sample_name)

    logger.info(f"{_dtnow()}: Preprocess {arguments['sample']}...")

    for path_fasta in Path(ARGS.tempdir, ARGS.control_name, "fasta").glob("*.fasta"):
        shutil.copy(path_fasta, Path(ARGS.tempdir, ARGS.sample_name, "fasta"))
    # ============================================================
    # Mapping with mappy
    # ============================================================
    paths_fasta = Path(ARGS.tempdir, ARGS.sample_name, "fasta").glob("*.fasta")
    preprocess.mapping.generate_sam(ARGS.tempdir, paths_fasta, ARGS.path_sample, ARGS.sample_name, ARGS.threads)
    # ============================================================
    # MIDSV conversion
    # ============================================================
    preprocess.midsv_caller.execute(ARGS.tempdir, ARGS.fasta_alleles, ARGS.sample_name)
    # ============================================================
    # Extract mutation loci
    # ============================================================
    preprocess.extract_knockin_loci(ARGS.tempdir, ARGS.sample_name)
    preprocess.extract_mutation_loci(ARGS.tempdir, ARGS.fasta_alleles, ARGS.sample_name, ARGS.control_name)
    # ============================================================
    # Detect and align insertion alleles
    # ============================================================
    paths_predifined_allele = {str(p) for p in Path(ARGS.tempdir, ARGS.sample_name, "fasta").glob("*.fasta")}
    preprocess.generate_insertion_fasta(ARGS.tempdir, ARGS.sample_name, ARGS.control_name, ARGS.fasta_alleles)
    paths_insertion = {str(p) for p in Path(ARGS.tempdir, ARGS.sample_name, "fasta").glob("insertion*.fasta")}
    paths_insertion -= paths_predifined_allele
    if paths_insertion:
        # mapping to insertion alleles
        preprocess.mapping.generate_sam(
            ARGS.tempdir, paths_insertion, ARGS.path_control, ARGS.control_name, ARGS.threads
        )
        preprocess.mapping.generate_sam(ARGS.tempdir, paths_insertion, ARGS.path_sample, ARGS.sample_name, ARGS.threads)
        # add insertions to ARGS.fasta_alleles
        for path_fasta in paths_insertion:
            allele, seq = Path(path_fasta).read_text().strip().split("\n")
            allele = allele.replace(">", "")
            ARGS.fasta_alleles[allele] = seq
        # MIDSV conversion
        preprocess.midsv_caller.execute(ARGS.tempdir, ARGS.fasta_alleles, ARGS.control_name)
        preprocess.midsv_caller.execute(ARGS.tempdir, ARGS.fasta_alleles, ARGS.sample_name)
        # Reculculate mutation loci
        preprocess.extract_mutation_loci(
            ARGS.tempdir, ARGS.fasta_alleles, ARGS.sample_name, ARGS.control_name, is_control=True
        )
        preprocess.extract_knockin_loci(ARGS.tempdir, ARGS.sample_name)
        preprocess.extract_mutation_loci(ARGS.tempdir, ARGS.fasta_alleles, ARGS.sample_name, ARGS.control_name)
    ########################################################################
    # Classify alleles
    ########################################################################
    logger.info(f"{_dtnow()}: Classify {arguments['sample']}...")
    classif_sample = classification.classify_alleles(ARGS.tempdir, ARGS.fasta_alleles, ARGS.sample_name)
    with open(Path(ARGS.tempdir, ARGS.sample_name, "classif_sample.pickle"), "wb") as p:
        pickle.dump(classif_sample, p)
    ########################################################################
    # Clustering
    ########################################################################
    logger.info(f"{_dtnow()}: Clustering {arguments['sample']}...")

    clust_sample = clustering.add_labels(classif_sample, ARGS.tempdir, ARGS.sample_name, ARGS.control_name)
    clust_sample = clustering.add_readnum(clust_sample)
    clust_sample = clustering.add_percent(clust_sample)
    clust_sample = clustering.update_labels(clust_sample)

    with open(Path(ARGS.tempdir, ARGS.sample_name, "clust_sample.pickle"), "wb") as p:
        pickle.dump(clust_sample, p)

    ########################################################################
    # Consensus call
    ########################################################################

    logger.info(f"{_dtnow()}: Consensus calling of {arguments['sample']}...")

    # Downsampling to 1000 reads in each LABEL
    clust_subset_sample = consensus.subset_clust(clust_sample, 1000)
    cons_percentage, cons_sequence = consensus.call_consensus(ARGS.tempdir, ARGS.sample_name, clust_subset_sample)
    allele_names = consensus.call_allele_name(cons_sequence, cons_percentage, ARGS.fasta_alleles)
    cons_percentage = consensus.update_key_by_allele_name(cons_percentage, allele_names)
    cons_sequence = consensus.update_key_by_allele_name(cons_sequence, allele_names)
    RESULT_SAMPLE = consensus.add_key_by_allele_name(clust_sample, allele_names)
    RESULT_SAMPLE.sort(key=lambda x: x["LABEL"])

    ########################################################################
    # Output Report：RESULT/FASTA/HTML/BAM
    ########################################################################

    logger.info(f"{_dtnow()}: Output reports of {arguments['sample']}...")

    # RESULT
    io.write_jsonl(RESULT_SAMPLE, Path(ARGS.tempdir, "result", f"{ARGS.sample_name}.jsonl"))
    # FASTA
    report.report_files.to_fasta(ARGS.tempdir, ARGS.sample_name, cons_sequence)
    # HTML
    report.report_files.to_html(ARGS.tempdir, ARGS.sample_name, cons_percentage)
    # CSV (Allele Info)
    report.report_mutation.to_csv(ARGS.tempdir, ARGS.sample_name, ARGS.genome_coordinates, cons_percentage)
    # BAM
    report.report_bam.output_bam(ARGS.tempdir, ARGS.sample_name, ARGS.genome_coordinates, ARGS.threads, RESULT_SAMPLE)
    for path_bam_igvjs in Path(ARGS.tempdir, "cache", ".igvjs").glob(f"{ARGS.control_name}_control.bam*"):
        shutil.copy(path_bam_igvjs, Path(ARGS.tempdir, "report", ".igvjs", ARGS.sample_name))
    # VCF
    # working in progress
    ###########################################################
    # Finish call
    ###########################################################
    logger.info(f"{_dtnow()}: \N{teacup without handle} {arguments['sample']} is finished!")



from DAJIN2.core.clustering.clustering import add_labels
from DAJIN2.core.clustering.clustering import add_percent
from DAJIN2.core.clustering.clustering import add_readnum
from DAJIN2.core.clustering.clustering import update_labels



from itertools import groupby


def subset_clust(clust_sample, num: int = 1000):
    clust_subset_sample = []
    clust_sample.sort(key=lambda x: x["LABEL"])
    for _, group in groupby(clust_sample, key=lambda x: x["LABEL"]):
        clust_subset_sample.extend(list(group)[:num])
    return clust_subset_sample



import os

# import psutil
import resource

# limit max memory usage
# available_memory = psutil.virtual_memory().available
available_memory = os.sysconf("SC_PAGE_SIZE") * os.sysconf("SC_PHYS_PAGES")
_, hard_limit = resource.getrlimit(resource.RLIMIT_DATA)
# if hard_limit == resource.RLIM_INFINITY:
#     hard_limit = available_memory
# Ensure that the new limit does not exceed the hard limit
limit = min(int(available_memory * 0.9), hard_limit)
resource.setrlimit(resource.RLIMIT_DATA, (limit, limit))



from __future__ import annotations

import re
from collections import defaultdict
import json
import numpy as np
from pathlib import Path
from typing import Generator
from scipy import stats
from scipy.spatial import distance
import pickle

# from sklearn.neighbors import LocalOutlierFactor
from DAJIN2.core.preprocess.homopolymer_handler import extract_errors_in_homopolymer


def read_midsv(filepath) -> Generator[dict[str, str]]:
    with open(filepath, "r") as f:
        for line in f:
            yield json.loads(line)


def call_coverage_on_each_base(midsv_sample: Generator[dict], sequence: str, index_mapping: dict) -> list[int]:
    coverages = [1] * len(sequence)
    for samp in midsv_sample:
        for i, cs in enumerate(samp["CSSPLIT"].split(",")):
            if i not in index_mapping:
                continue
            if cs == "N":
                continue
            coverages[index_mapping[i]] += 1
    return coverages


def count_indels(midsv_sample, sequence: str, index_mapping: dict) -> dict[str, list[int]]:
    len_sequence = len(sequence)
    count = {"+": [0] * len_sequence, "-": [0] * len_sequence, "*": [0] * len_sequence}
    for samp in midsv_sample:
        for i, cs in enumerate(samp["CSSPLIT"].split(",")):
            if i not in index_mapping:
                continue
            if cs.startswith("=") or cs == "N" or re.search(r"a|c|g|t|n", cs):
                continue
            if cs.startswith("+"):
                # count["+"][i] += len(cs.split("|"))
                count["+"][i] += 1
            elif cs.startswith("-"):
                count["-"][i] += 1
            elif cs.startswith("*"):
                count["*"][i] += 1
    return count


def normalize_indels(count: dict[str, list[int]], coverages: list[int]) -> dict[str, np.array]:
    count_normalized = dict()
    coverages = np.array(coverages)
    for mut in count:
        counts = np.array(count[mut])
        count_normalized[mut] = counts / coverages
    return count_normalized


def split_kmer(indels: dict[str, np.array], kmer: int = 11) -> dict[str, np.array]:
    results = defaultdict(list)
    center = kmer // 2
    for mut, value in indels.items():
        for i in range(len(value)):
            if center <= i <= len(value) - center:
                start = i - center
                if kmer % 2 == 0:
                    end = i + center
                else:
                    end = i + center + 1
                results[mut].append(value[start:end])
            else:
                results[mut].append(np.array([0] * kmer))
    return results


def merge_peaks(indels_sample_normalized, mutation, peaks, hight):
    """Values higher than 75% quantile of the control values and the surrouings are peaks, merge it as a peak"""
    values_sample = indels_sample_normalized[mutation]
    for i, value in enumerate(values_sample):
        if i not in peaks and value > hight:
            for j in range(i - 5, i + 6):
                if j in peaks:
                    peaks.add(i)
                    break
    return peaks


def extract_dissimilar_loci(indels_kmer_sample: dict, indels_kmer_control: dict) -> dict[str, set]:
    """
    Comparing Sample and Control, the 'similar mean' and
    'similar variance' are considered as sequence errors.
    """
    results = dict()
    for mut in indels_kmer_sample:
        values_sample = indels_kmer_sample[mut]
        values_control = indels_kmer_control[mut]
        """
        Calculate cosine similarity: 1 means exactly same, 0 means completely different.
        Zero vector does not return correct value, so add epsilon (1e-10).
        example: distance.cosine([0,0,0], [1,2,3]) retuns 0.
        """
        cossims = [1 - distance.cosine(x + 1e-10, y + 1e-10) for x, y in zip(values_sample, values_control)]
        # Perform T-test: nan means exactly same, p > 0.05 means similar in average.
        t_pvalues = [stats.ttest_ind(x, y, equal_var=False)[1] for x, y in zip(values_sample, values_control)]
        t_pvalues = [1 if np.isnan(t) else t for t in t_pvalues]
        # Perform F-test: p > 0.05 means similar in variance.
        f_pvalues = [stats.bartlett(x, y)[1] for x, y in zip(values_sample, values_control)]
        # if pvalue == nan or pval > 0.05, samples and controls are similar.
        dissimilar_loci = set()
        for i, (cossim, t_pval, f_pval) in enumerate(zip(cossims, t_pvalues, f_pvalues)):
            # deletion may be the similar cossim, but t_val < 0.05. Others should be dissimilar.
            if (cossim > 0.9 and t_pval < 0.05) or (cossim < 0.9 and (t_pval < 0.05 or f_pval < 0.05)):
                dissimilar_loci.add(i)
        results[mut] = dissimilar_loci
    return results


def discard_errors_in_homopolymer(dissimilar_loci, errors_in_homopolymer) -> dict[str, set]:
    mutation_loci = dict()
    for mut in ["+", "-", "*"]:
        error_loci = errors_in_homopolymer[mut]
        mutation_loci[mut] = dissimilar_loci[mut] - error_loci
    return mutation_loci


def transpose_mutation_loci(mutation_loci: set[int], sequence: str) -> list[set]:
    mutation_loci_transposed = [set() for _ in range(len(sequence))]
    for mut, idx_mutation in mutation_loci.items():
        for i, loci in enumerate(mutation_loci_transposed):
            if i in idx_mutation:
                loci.add(mut)
    return mutation_loci_transposed


###########################################################
# main
###########################################################


def process_mutation_loci(TEMPDIR: Path, FASTA_ALLELES: dict, CONTROL_NAME: str) -> None:
    with open(Path(TEMPDIR, "mutation_loci", "index_mapping..pickle"), "rb") as f:
        INDEX_MAPPING = pickle.load(f)
    for allele, sequence in FASTA_ALLELES.items():
        index_mapping = INDEX_MAPPING.get(allele, {i: i for i in range(len(sequence))})
        index_mapping = {v: k for k, v in index_mapping.items()}
        filepath_control = Path(TEMPDIR, "midsv", f"{CONTROL_NAME}_{allele}.json")
        indels_control = count_indels(read_midsv(filepath_control), sequence, index_mapping)
        coverages_control = call_coverage_on_each_base(read_midsv(filepath_control), sequence, index_mapping)
        indels_control_normalized = normalize_indels(indels_control, coverages_control)
        indels_kmer_control = split_kmer(indels_control_normalized, kmer=11)
        # Save indels_control_normalized and indels_kmer_control as pickle to reuse in consensus calling
        with open(Path(TEMPDIR, "mutation_loci", f"{CONTROL_NAME}_{allele}_normalized..pickle"), "wb") as f:
            pickle.dump(indels_control_normalized, f)
        with open(Path(TEMPDIR, "mutation_loci", f"{CONTROL_NAME}_{allele}_kmer..pickle"), "wb") as f:
            pickle.dump(indels_kmer_control, f)


def update_mutation_loci(MUTATION_LOCI_ALLELES, INDEX_MAPPING) -> dict[str, list]:
    """
    Filter mutations that exists both in control and other alleles
    """
    mutation_loci_control = MUTATION_LOCI_ALLELES["control"]
    for allele in MUTATION_LOCI_ALLELES:
        if allele == "control":
            continue
        index_mapping = INDEX_MAPPING[allele]
        for i, mut_control in enumerate(mutation_loci_control):
            corresponded_mutation = mut_control & MUTATION_LOCI_ALLELES[allele][index_mapping[i]]
            if corresponded_mutation:
                MUTATION_LOCI_ALLELES[allele][i] = corresponded_mutation
            else:
                MUTATION_LOCI_ALLELES["control"][i] = set()
                MUTATION_LOCI_ALLELES[allele][index_mapping[i]] = set()
    return MUTATION_LOCI_ALLELES


def extract_mutation_loci(TEMPDIR: Path, FASTA_ALLELES: dict, SAMPLE_NAME: str, CONTROL_NAME: str) -> dict[str, list]:
    with open(Path(TEMPDIR, "mutation_loci", "index_mapping..pickle"), "rb") as f:
        INDEX_MAPPING = pickle.load(f)
    MUTATION_LOCI_ALLELES = dict()
    for allele, sequence in FASTA_ALLELES.items():
        filepath_sample = Path(TEMPDIR, "midsv", f"{SAMPLE_NAME}_{allele}.json")
        index_mapping = INDEX_MAPPING.get(allele, {i: i for i in range(len(sequence))})
        index_mapping = {v: k for k, v in index_mapping.items()}
        indels_sample = count_indels(read_midsv(filepath_sample), sequence, index_mapping)
        coverages_sample = call_coverage_on_each_base(read_midsv(filepath_sample), sequence, index_mapping)
        indels_sample_normalized = normalize_indels(indels_sample, coverages_sample)
        indels_kmer_sample = split_kmer(indels_sample_normalized, kmer=11)
        # Load indels_control_normalized and indels_kmer_control
        with open(Path(TEMPDIR, "mutation_loci", f"{CONTROL_NAME}_{allele}_normalized..pickle"), "rb") as f:
            indels_control_normalized = pickle.load(f)
        with open(Path(TEMPDIR, "mutation_loci", f"{CONTROL_NAME}_{allele}_kmer..pickle"), "rb") as f:
            indels_kmer_control = pickle.load(f)
        # Calculate dissimilar loci
        dissimilar_loci = extract_dissimilar_loci(indels_kmer_sample, indels_kmer_control)
        # Extract error loci in homopolymer regions
        errors_in_homopolymer = dict()
        for mut in ["+", "-", "*"]:
            indels_sample_mut = indels_sample_normalized[mut]
            indels_control_mut = indels_control_normalized[mut]
            candidate_loci = dissimilar_loci[mut]
            # candidate_loci = anomaly_loci[mut] & dissimilar_loci[mut]
            errors_in_homopolymer[mut] = extract_errors_in_homopolymer(
                indels_sample_mut, indels_control_mut, sequence, candidate_loci
            )
        mutation_loci = discard_errors_in_homopolymer(dissimilar_loci, errors_in_homopolymer)
        mutation_loci_transposed = transpose_mutation_loci(mutation_loci, sequence)
        MUTATION_LOCI_ALLELES[allele] = mutation_loci_transposed
    MUTATION_LOCI_ALLELES = update_mutation_loci(MUTATION_LOCI_ALLELES, INDEX_MAPPING)
    return MUTATION_LOCI_ALLELES



from __future__ import annotations

import os
import sys
import logging
import datetime
import argparse

import traceback
from multiprocessing import Process, Queue

from pathlib import Path
from itertools import groupby, islice
from typing import Generator

from DAJIN2 import gui, view
from DAJIN2.core import core
from DAJIN2.utils import config, io, report_generator, input_validator


VERSION = "0.3.1"

# prevent BLAS from using all cores
os.environ["OMP_NUM_THREADS"] = "1"
os.environ["OPENBLAS_NUM_THREADS"] = "1"
os.environ["MKL_NUM_THREADS"] = "1"
os.environ["VECLIB_MAXIMUM_THREADS"] = "1"
os.environ["NUMEXPR_NUM_THREADS"] = "1"

###########################################################
# Setting logger
###########################################################

current_time = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")

stderr_handler = logging.StreamHandler()
stderr_handler.setFormatter(logging.Formatter("%(message)s"))

file_handler = logging.FileHandler(f"{current_time}_DAJIN2.log")
file_handler.setFormatter(logging.Formatter("%(message)s"))

logging.basicConfig(
    level=logging.INFO,
    format="[%(asctime)s] [%(levelname)s] [%(processName)s] %(message)s",
    handlers=[stderr_handler, file_handler],
)


def update_threads(threads: int) -> int:
    available_threads = os.cpu_count() - 1
    threads_updated = max(1, min(threads, available_threads))
    return threads_updated


def generate_report(name: str) -> None:
    report_generator.report(name)
    print(
        f"\N{party popper} Finished! Open {config.DAJIN_RESULTS_DIR}/{name} to see the report.",
        file=sys.stderr,
    )


################################################################################
# Single mode
################################################################################


def execute_single_mode(arguments: dict[str]):
    input_validator.validate_files(arguments["sample"], arguments["control"], arguments["allele"])
    if "genome" in arguments:
        arguments.update(input_validator.validate_genome_and_fetch_urls(arguments["genome"]))
    core.execute_control(arguments)
    core.execute_sample(arguments)
    generate_report(arguments["name"])


################################################################################
# Batch mode
################################################################################


def validate_columns(columns: list, filepath: str) -> None:
    """Validate the columns of a batch file."""
    required_columns = ["sample", "control", "allele", "name"]
    accepted_columns = ["sample", "control", "allele", "name", "genome"]

    if not set(required_columns).issubset(set(columns)):
        raise ValueError(f"{filepath} must contain {', '.join(required_columns)} in the header")

    if not set(columns).issubset(accepted_columns):
        raise ValueError(f"Accepted header names of {filepath} are {', '.join(accepted_columns)}.")


def _batched(iterable, chunk_size: int) -> Generator(tuple):
    iterator = iter(iterable)
    while True:
        chunk = tuple(islice(iterator, chunk_size))
        if not chunk:
            return
        yield chunk


def run_multiprocess(function, arguments: list[dict], num_workers: int = 1) -> None:
    logger = logging.getLogger(__name__)

    q = Queue()

    def target(arg):
        try:
            function(arg)
        except Exception:
            q.put(traceback.format_exc())
            sys.exit(1)

    arguments_batched = _batched(arguments, num_workers)
    for args in arguments_batched:
        processes = [Process(target=target, args=(arg,)) for arg in args]

        for p in processes:
            p.start()

        for p in processes:
            p.join()
            if p.exitcode == 1:
                logger.error(f"An unexpected error occurred: {q.get()}")
                sys.exit(1)


def execute_batch_mode(arguments: dict[str]):
    path_batchfile = io.convert_to_posix(arguments["file"])

    if not Path(path_batchfile).exists():
        raise FileNotFoundError(f"'{path_batchfile}' does not exist.")

    inputs = io.load_file(path_batchfile)

    # Validate Column of the batch file
    columns = inputs[0]
    validate_columns(columns, path_batchfile)

    # Validate contents and fetch genome urls
    contents = inputs[1:]
    cache_urls_genome = dict()
    index_of_name = columns.index("name")
    contents.sort(key=lambda x: x[index_of_name])
    for _, groups in groupby(contents, key=lambda x: x[index_of_name]):
        for group in groups:
            args = {h: g for h, g in zip(columns, group)}
            # validate contents in the batch file
            input_validator.validate_files(args["sample"], args["control"], args["allele"])
            # validate genome and fetch urls
            if "genome" in args and args["genome"] not in cache_urls_genome:
                urls_genome = input_validator.validate_genome_and_fetch_urls(args["genome"])
                cache_urls_genome[args["genome"]] = urls_genome

    ##############################################################################
    # Perform DAJIN2
    ##############################################################################

    index_of_name = columns.index("name")
    contents.sort(key=lambda x: x[index_of_name])
    contents_sample = []
    for name, groups in groupby(contents, key=lambda x: x[index_of_name]):
        groups = list(groups)

        # Handle controls
        contents_control = []
        for group in groups:
            args = {h: g for h, g in zip(columns, group)}
            args["threads"] = arguments["threads"]
            args["sample"] = args["control"]
            if "genome" in args:
                args.update(cache_urls_genome[args["genome"]])
            contents_control.append(args)
        contents_control_unique = [dict(item) for item in set(frozenset(d.items()) for d in contents_control)]
        run_multiprocess(core.execute_control, contents_control_unique, arguments["threads"])

        # Handle samples
        contents_sample = []
        for group in groups:
            args = {h: g for h, g in zip(columns, group)}
            args["threads"] = arguments["threads"]
            if args["sample"] == args["control"]:
                continue
            if "genome" in args:
                args.update(cache_urls_genome[args["genome"]])
            contents_sample.append(args)
        contents_sample_unique = [dict(item) for item in set(frozenset(d.items()) for d in contents_sample)]
        run_multiprocess(core.execute_sample, contents_sample_unique, arguments["threads"])
        # Finish
        generate_report(name)


def execute():
    parser = argparse.ArgumentParser()

    ###############################################################################
    # Single mode
    ###############################################################################

    parser.add_argument("-s", "--sample", type=str, help="Full path to a sample FASTQ file")
    parser.add_argument("-c", "--control", type=str, help="Full path to a control FASTQ file")
    parser.add_argument("-a", "--allele", type=str, help="Full path to a FASTA file")
    parser.add_argument("-n", "--name", type=str, help="Output directory name", default="DAJIN2-results")
    parser.add_argument(
        "-g", "--genome", type=str, default="", help="Reference genome ID (e.g hg38, mm39) [default: '']"
    )
    parser.add_argument("-t", "--threads", type=int, default=1, help="Number of threads [default: 1]")
    parser.add_argument("-v", "--version", action="version", version=f"DAJIN2 version {VERSION}")
    parser.add_argument("-d", "--debug", action="store_true", help=argparse.SUPPRESS)

    ###############################################################################
    # Batch mode
    ###############################################################################

    def batchmode(args):
        arguments = dict()
        arguments["file"] = args.file
        arguments["threads"] = update_threads(int(args.threads))
        arguments["debug"] = args.debug
        execute_batch_mode(arguments)

    subparser = parser.add_subparsers()
    parser_batch = subparser.add_parser("batch", help="DAIJN2 batch mode")
    parser_batch.add_argument("-f", "--file", required=True, type=str, help="CSV or Excel file.")
    parser_batch.add_argument("-t", "--threads", default=1, type=int, help="Number of threads [default: 1]")
    parser_batch.add_argument("-d", "--debug", action="store_true", help=argparse.SUPPRESS)
    parser_batch.set_defaults(handler=batchmode)

    ###############################################################################
    # GUI mode
    ###############################################################################

    def guimode(args):
        gui.execute()

    parser_gui = subparser.add_parser("gui", help="DAIJN2 GUI mode")
    parser_gui.set_defaults(handler=guimode)

    ###############################################################################
    # View mode
    ###############################################################################

    def viewmode(args):
        view.execute(args.name)

    parser_view = subparser.add_parser("view", help="DAIJN2 View mode to launch igvjs")
    parser_view.add_argument("-n", "--name", required=True, type=str, help="Output name of the report")
    parser_view.set_defaults(handler=viewmode)

    ###############################################################################
    # Parse arguments
    ###############################################################################

    args = parser.parse_args()

    if hasattr(args, "handler"):
        args.handler(args)
    else:
        if args.sample is None:
            raise ValueError("the following arguments are required: -s/--sample")
        if args.control is None:
            raise ValueError("the following arguments are required: -c/--control")
        if args.allele is None:
            raise ValueError("the following arguments are required: -a/--allele")
        if args.name is None:
            raise ValueError("the following arguments are required: -n/--name")
        arguments = dict()
        arguments["sample"] = args.sample
        arguments["control"] = args.control
        arguments["allele"] = args.allele
        arguments["name"] = args.name
        arguments["genome"] = args.genome
        arguments["threads"] = update_threads(int(args.threads))
        arguments["debug"] = args.debug

        logger = logging.getLogger(__name__)
        try:
            execute_single_mode(arguments)
        except Exception as e:
            logger.error(f"An unexpected error occurred: {e}")
            sys.exit(1)


if __name__ == "__main__":
    execute()



from __future__ import annotations

import json
import warnings
import numpy as np

from pathlib import Path
from itertools import chain
from typing import Generator
from collections import Counter
from collections import defaultdict

from sklearn import metrics
from sklearn.decomposition import PCA
from sklearn.mixture import GaussianMixture
from sklearn.tree import DecisionTreeClassifier
from sklearn.exceptions import ConvergenceWarning

from DAJIN2.core.clustering.merge_clusters import merge_clusters

warnings.filterwarnings("ignore", category=RuntimeWarning)
warnings.filterwarnings("ignore", category=ConvergenceWarning)

###############################################################################
# I/O
###############################################################################


def read_json(filepath: Path | str) -> Generator[list[float]]:
    with open(filepath, "r") as f:
        for line in f:
            yield json.loads(line)


def count_newlines(filepath: Path | str) -> int:
    def _make_gen(reader):
        while True:
            b = reader(2**16)
            if not b:
                break
            yield b

    with open(filepath, "rb") as f:
        count = sum(buf.count(b"\n") for buf in _make_gen(f.raw.read))
    return count


###############################################################################
# Dimension reduction
###############################################################################


def reduce_dimension(scores_sample: Generator[list], scores_control: Generator[list]) -> np.array:
    scores = list(chain(scores_sample, scores_control))
    model = PCA(n_components=min(20, len(scores)))
    # return model.fit_transform(scores) * np.nan_to_num(model.explained_variance_ratio_)
    return model.fit_transform(scores)


###############################################################################
# Clustering
###############################################################################


def optimize_labels(X: np.array, coverage_sample, coverage_control) -> list[int]:
    n_components = min(20, coverage_sample + coverage_control)
    labels_prev = list(range(coverage_sample))
    for i in range(1, n_components):
        np.random.seed(seed=1)
        labels_all = GaussianMixture(n_components=i, random_state=1).fit_predict(X).tolist()
        labels_sample = labels_all[:coverage_sample]
        labels_control = labels_all[coverage_sample:]
        labels_merged = merge_clusters(labels_control, labels_sample)
        # print(i, Counter(labels_sample), Counter(labels_control), Counter(labels_merged))  # ! DEBUG
        # Reads < 1% in the control are considered clustering errors and are not counted
        count_control = Counter(labels_control)
        num_labels_control = sum(1 for reads in count_control.values() if reads / coverage_control > 0.01)
        mutual_info = metrics.adjusted_mutual_info_score(labels_prev, labels_merged)
        # Report the number of clusters in SAMPLE when the number of clusters in CONTROL is split into more than one.
        if num_labels_control > 1 or 0.95 < mutual_info < 1.0:
            return labels_merged
        else:
            labels_results = labels_merged
        labels_prev = labels_merged
    return labels_results


###############################################################################
# Handle Strand bias
# # Clusters of reads with mutations with strand bias are merged into similar clusters without strand bias
###############################################################################


def _count_strand(labels, samples) -> tuple(defaultdict, defaultdict):
    count_strand_by_labels = defaultdict(int)
    total_count_by_labels = defaultdict(int)
    for label, sample in zip(labels, samples):
        total_count_by_labels[label] += 1
        if sample["STRAND"] == "+":
            count_strand_by_labels[label] += 1
    return count_strand_by_labels, total_count_by_labels


def _calculate_strand_biases(count_strand_by_labels, total_count_by_labels) -> dict[int, bool]:
    strand_biases = dict()
    for label, total in total_count_by_labels.items():
        strand = count_strand_by_labels[label]
        strand_ratio = strand / total
        strand_biases[label] = not (0.25 < strand_ratio < 0.75)
    return strand_biases


def get_strand_biases(labels: list[int], path_sample: Path | str) -> dict[int, bool]:
    samples = read_json(path_sample)
    count_strand_by_labels, total_count_by_labels = _count_strand(labels, samples)
    return _calculate_strand_biases(count_strand_by_labels, total_count_by_labels)


def _prepare_training_testing_sets(labels, scores, strand_biases):
    x_train, y_train, x_test = [], [], []
    for label, score in zip(labels, scores):
        if strand_biases[label]:
            x_test.append(score)
        else:
            x_train.append(score)
            y_train.append(label)
    return x_train, y_train, x_test


def _train_decision_tree(x_train, y_train):
    dtree = DecisionTreeClassifier(random_state=1)
    dtree.fit(x_train, y_train)
    return dtree


def _allocate_labels(labels, strand_biases, dtree, x_test):
    label_predictions = dtree.predict(x_test)
    label_predict_iter = iter(label_predictions)
    for i, label in enumerate(labels):
        if strand_biases[label]:
            labels[i] = next(label_predict_iter)
    return labels


def correct_clusters_with_strand_bias(labels, path_score_sample, strand_biases):
    scores = read_json(path_score_sample)
    x_train, y_train, x_test = _prepare_training_testing_sets(labels, scores, strand_biases)
    dtree = _train_decision_tree(x_train, y_train)
    return _allocate_labels(labels, strand_biases, dtree, x_test)


###############################################################################
# main
###############################################################################


def get_label_most(labels: list[int]) -> int:
    return Counter(labels).most_common()[0][0]


def subset_scores(labels: list[int], scores: list[int], label_most: int, length: int = 1000) -> list[int]:
    subset = [score for label, score in zip(labels, scores) if label == label_most]
    return subset[:length]


def return_labels(
    path_score_sample: Path | str, path_score_control: Path | str, path_sample: Path | str, strand_bias: bool
) -> list[int]:
    np.random.seed(seed=1)
    X_control = reduce_dimension([], read_json(path_score_control))
    # subset to 1000 reads of controls in the most common cluster to remove outliers and reduce computation time
    labels_control = GaussianMixture(n_components=2, random_state=1).fit_predict(X_control)
    label_most = get_label_most(labels_control)
    scores_control_subset = subset_scores(labels_control, read_json(path_score_control), label_most, 1000)
    X = reduce_dimension(read_json(path_score_sample), scores_control_subset)
    coverage_sample = count_newlines(path_score_sample)
    coverage_control = len(scores_control_subset)
    labels = optimize_labels(X, coverage_sample, coverage_control)
    # correct clusters with strand bias
    if strand_bias is False:
        strand_biases = get_strand_biases(labels, path_sample)
        # Until there is at least one True and one False or
        # 1000 iterations (1000 is a suitable number to exit an infinite loop just in case)
        i = 0
        while len(Counter(strand_biases.values())) > 1 and i < 1000:
            labels = correct_clusters_with_strand_bias(labels, path_score_sample, strand_biases)
            strand_biases = get_strand_biases(labels, path_sample)
            i += 1
    return labels



from __future__ import annotations

from typing import Generator
from itertools import groupby
from collections import Counter

from DAJIN2.core.clustering.make_kmer import generate_mutation_kmers


def _call_count(cssplits_sample: Generator[list[str]]) -> list[dict[str, int]]:
    count_kmer = []
    for cssplits_transposed in zip(*cssplits_sample):
        count_kmer.append(dict(Counter(cssplits_transposed)))
    return count_kmer


def _call_percent(counts: list[dict[str, int]]) -> list[dict[str, float]]:
    cssplit_percent = []
    coverage = sum(counts[0].values())
    for count in counts:
        percent = {k: round(v / coverage * 100, 3) for k, v in count.items()}
        cssplit_percent.append(percent)
    return cssplit_percent


def _subtract_percentage(percent_control, percent_sample, knockin_loci: set(int)) -> list[dict]:
    sample_subtracted = []
    for i, (cont, samp) in enumerate(zip(percent_control, percent_sample)):
        if i in knockin_loci:
            sample_subtracted.append(dict(samp))
            continue
        samp = Counter(samp)
        samp.subtract(Counter(cont))
        sample_subtracted.append(dict(samp))
    return sample_subtracted


def _discard_common_error(percent_subtraction, threshold=0.5) -> list[dict]:
    percent_discarded = []
    for samp in percent_subtraction:
        remained = {k: v for k, v in samp.items() if v > threshold}
        percent_discarded.append(remained)
    return percent_discarded


def _discard_match_and_n(percent_discarded) -> list[dict]:
    mutation_score_discarded = [dict() for _ in range(len(percent_discarded))]
    for i, mutation_percent in enumerate(percent_discarded):
        if mutation_percent == {} or "" in mutation_percent:
            continue
        for mutation, percent in mutation_percent.items():
            mutation_center = mutation.split(",")[1]
            if mutation_center.startswith("=") or mutation_center == ("N"):
                continue
            mutation_score_discarded[i].update({mutation: percent})
    return mutation_score_discarded


###############################################################################
# Handling insertions
###############################################################################


def _group_consecutive_insertions(mutation_loci: dict[str, set[int]]) -> list[tuple[int]]:
    index = sorted(i for i, m in enumerate(mutation_loci) if "+" in m)
    index_grouped = []
    for _, group in groupby(enumerate(index), lambda i_x: i_x[0] - i_x[1]):
        items = [value for _, value in group]
        index_grouped.append(tuple(items))
    return index_grouped


def update_insertion_score(percent_discarded, mutation_loci):
    index_grouped = _group_consecutive_insertions(mutation_loci)
    for index_insertion in index_grouped:
        max_score = 0
        for i in index_insertion:
            for key, val in percent_discarded[i].items():
                if key.split(",")[1].startswith("+"):
                    max_score = max(max_score, val)
        for i in index_insertion:
            for key in percent_discarded[i]:
                if key.split(",")[1].startswith("+"):
                    percent_discarded[i][key] = max_score
    return percent_discarded


###############################################################################
# main
###############################################################################


def make_score(path_sample, path_control, mutation_loci: set[int], knockin_loci: set[int]) -> list[dict[str, float]]:
    cssplits_sample = generate_mutation_kmers(path_sample, mutation_loci, compress_ins=True)
    cssplits_control = generate_mutation_kmers(path_control, mutation_loci, compress_ins=True)
    counts_sample = _call_count(cssplits_sample)
    counts_control = _call_count(cssplits_control)
    percent_sample = _call_percent(counts_sample)
    percent_control = _call_percent(counts_control)
    percent_subtraction = _subtract_percentage(percent_control, percent_sample, knockin_loci)
    percent_discarded = _discard_common_error(percent_subtraction, 0.5)
    percent_discarded = update_insertion_score(percent_discarded, mutation_loci)
    mutation_score = _discard_match_and_n(percent_discarded)
    return mutation_score



from __future__ import annotations

import os
import csv
import json
import hashlib

import wslPath
import pandas as pd

from pathlib import Path
from typing import Generator


###########################################################
# convert path to POSIX format
###########################################################


def convert_to_posix(path: str) -> str:
    if wslPath.is_windows_path(path):
        path = wslPath.to_posix(path)
    return path


###########################################################
# Load batch file
###########################################################


def check_excel_or_csv(filepath: str) -> str:
    """Check if the file is an Excel or CSV file. Raise error for other types."""
    _, ext = os.path.splitext(filepath)
    if ext in [".xlsx", ".xls"]:
        return "excel"
    elif ext == ".csv":
        return "csv"
    else:
        raise ValueError("The provided file must be either an Excel or CSV file.")


def load_from_excel(path: str) -> list:
    """Load data from an Excel file and return as a list."""
    df = pd.read_excel(path)
    return [df.columns.to_list()] + df.values.tolist()


def load_from_csv(path: str) -> list:
    """Load data from a CSV file and return as a list."""
    with open(path, "r") as f:
        return [row for row in csv.reader(f, skipinitialspace=True, delimiter=",")]


def load_file(path_batchfile: str) -> list:
    """Load data from either an Excel or CSV file."""
    file_type = check_excel_or_csv(path_batchfile)
    if file_type == "excel":
        return load_from_excel(path_batchfile)
    elif file_type == "csv":
        return load_from_csv(path_batchfile)


###########################################################
# Input/Output Json
###########################################################


def read_jsonl(path: str) -> Generator[dict]:
    with open(path, "r") as f:
        for line in f:
            yield json.loads(line)


def write_jsonl(data: list[dict], path: str) -> None:
    with open(path, "w") as f:
        for d in data:
            json_str = json.dumps(d)
            f.write(json_str + "\n")


###########################################################
# Cache control hash
###########################################################


def cache_control_hash(tempdir: Path, path_allele: str | Path) -> None:
    """
    Generate a hash value from the content of the provided file and cache it.

    Parameters:
    - tempdir: Path object, the temporary directory path.
    - path_allele: Path object, the path to the fasta file.
    """
    # Read the content of the file.
    content = Path(path_allele).read_text()
    # Calculate the hash of the content.
    content_hash = hashlib.sha256(content.encode()).hexdigest()
    # Define the path to save the hash.
    path_cache_hash = Path(tempdir, "cache", "hash.txt")
    # Ensure the directory exists.
    path_cache_hash.parent.mkdir(parents=True, exist_ok=True)
    # Save the hash to the file.
    path_cache_hash.write_text(content_hash)



import http.server
import os
import re
import socket
import socketserver
import webbrowser
from contextlib import closing, redirect_stderr
from pathlib import Path
from threading import Timer

from jinja2 import Environment, FileSystemLoader, Template

from DAJIN2.utils.config import DAJIN_RESULTS_DIR


def find_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def open_browser(PORT):
    webbrowser.open_new(f"http://127.0.0.1:{PORT}/")


def execute(name: str):
    DIR_IGVJS = Path(DAJIN_RESULTS_DIR, name, ".igvjs")
    if not DIR_IGVJS.exists():
        raise FileNotFoundError(f"BAM files are not found in {DIR_IGVJS.parent}. Please run DAJIN first.")
    path_view = Path(__file__).parent
    env = Environment(loader=FileSystemLoader(path_view, encoding="utf8"))
    template = env.get_template("template_igvjs.html")
    params_genome = {"genome": {"exist": False}}
    params_reference = dict()
    path_genome = Path(DIR_IGVJS, "genome_symbol.txt")
    if path_genome.exists():
        path_coodinates = Path(DIR_IGVJS, "genome_coodinates.jsonl")
        GENOME = path_genome.read_text().strip()
        CHROME, START, END, _ = eval(path_coodinates.read_text().strip()).values()
        params_genome = {"genome": {"exist": True, "genome": GENOME, "locus": f"{CHROME}:{START}-{END}"}}
    else:
        params_reference = {"reference": {"urlfa": "control.fasta", "urlfai": "control.fasta" + ".fai"}}

    bamnames = []
    bamurls = []
    baiurls = []
    for bam in Path(DAJIN_RESULTS_DIR, name, ".igvjs").iterdir():
        if not re.search(r"bam$", str(bam)):
            continue
        bamnames.append(bam.stem)
        bamurls.append(str(bam.name))
        baiurls.append(str(bam.name) + ".bai")
    bamnames.sort()
    bamurls.sort()
    baiurls.sort()
    contents = [{"samplename": n, "urlbam": b, "urlbai": i} for n, b, i in zip(bamnames, bamurls, baiurls)]

    params_tracks = """{
        "tracks":
            {{ contents }},
    }"""
    params_tracks = Template(params_tracks)
    params_tracks = eval(params_tracks.render(contents=contents))

    params_genome.update(**params_reference, **params_tracks)
    params = params_genome.copy()

    HTML_IGVJS = template.render(params)
    Path(DIR_IGVJS, "index.html").write_text(HTML_IGVJS)
    PORT = find_free_port()
    Handler = http.server.SimpleHTTPRequestHandler

    os.chdir(Path(DAJIN_RESULTS_DIR, name, ".igvjs"))
    with socketserver.TCPServer(("", PORT), Handler) as httpd:
        print(f"Assess 'http://127.0.0.1:{PORT}/' if a browser does not automatically open.")
        Timer(1, open_browser, [PORT]).start()
        with redirect_stderr(open(os.devnull, "w")):
            httpd.serve_forever()



from __future__ import annotations

from collections import Counter, defaultdict


def merge_mixed_cluster(labels_control: list[int], labels_sample: list[int]) -> list[int]:
    labels_merged = labels_sample.copy()
    coverage_control = len(labels_control)
    labels_percent_control = defaultdict(float)
    for i, label in enumerate(labels_control):
        labels_percent_control[label] += 1 / coverage_control * 100
    labels_mixed = {i for i, per in labels_percent_control.items() if per > 0.5}
    max_label = max(labels_merged) + 1
    for i, label in enumerate(labels_sample):
        if label in labels_mixed:
            labels_merged[i] = max_label
    return labels_merged


def merge_minor_cluster(labels_sample: list[int]) -> list[int]:
    total_num = sum(Counter(labels_sample).values())
    cnt = Counter(labels_sample)
    labels_minor = {label for label, num in cnt.items() if num / total_num * 100 < 0.5}
    comon_label = cnt.most_common()[0][0]
    labels_merged = labels_sample.copy()
    for i, label in enumerate(labels_sample):
        if label in labels_minor:
            labels_merged[i] = comon_label
    return labels_merged


###############################################################################
# main
###############################################################################


def merge_clusters(labels_control: list[int], labels_sample: list[int]) -> list[int]:
    labels_merged = merge_mixed_cluster(labels_control, labels_sample)
    labels_merged = merge_minor_cluster(labels_merged)
    return labels_merged



from __future__ import annotations

from urllib.request import urlopen


def _fetch_seq_coordinates(genome: str, blat_url: str, seq: str) -> dict:
    url = f"{blat_url}?db={genome}&type=BLAT&userSeq={seq}"
    response = urlopen(url).read().decode("utf8").split("\n")
    matches = [x for x in response if "100.0%" in x]
    if not matches:
        raise ValueError(f"{seq} is not found in {genome}")
    chrom, strand, start, end, _ = matches[0].split()[-5:]
    return {"chrom": chrom, "strand": strand, "start": int(start), "end": int(end)}


def fetch_coordinates(genome_coordinates: dict, genome_urls: dict, seq: str) -> dict:
    genome = genome_coordinates["genome"]
    blat_url = genome_urls["blat"]

    seq_start, seq_end = seq[:1000], seq[-1000:]
    coordinate_start = _fetch_seq_coordinates(genome, blat_url, seq_start)
    coordinate_end = _fetch_seq_coordinates(genome, blat_url, seq_end)

    chromosome, strand = coordinate_start["chrom"], coordinate_start["strand"]
    if strand == "+":
        start, end = coordinate_start["start"], coordinate_end["end"]
    else:
        start, end = coordinate_end["start"], coordinate_start["end"]

    return {"genome": genome, "chrom": chromosome, "start": start, "end": end, "strand": strand}


def fetch_chromosome_size(genome_coordinates: dict, genome_urls: dict) -> int:
    chrom = genome_coordinates["chrom"]
    genome = genome_coordinates["genome"]
    url = f"{genome_urls['goldenpath']}/{genome}/bigZips/{genome}.chrom.sizes"

    response = urlopen(url).read().decode("utf8").split("\n")
    for line in response:
        chrom_name, size = line.split("\t")
        if chrom == chrom_name:
            return int(size)
    raise ValueError(f"Chromosome {chrom} size not found.")



from __future__ import annotations

import hashlib
from pathlib import Path
from DAJIN2.utils import io


def exists_cached_hash(tempdir: Path, path: str) -> bool:
    path_cache_hash = Path(tempdir, "cache", "hash.txt")
    if path_cache_hash.exists():
        current_hash = hashlib.sha256(Path(path).read_bytes()).hexdigest()
        cashed_hash = path_cache_hash.read_text()
        if current_hash == cashed_hash:
            return True
    return False


def exists_cached_genome(tempdir: Path, genome: str) -> bool:
    path_cached_genome = Path(tempdir, "cache", "genome_coodinates.jsonl")
    if path_cached_genome.exists():
        cached_genome = list(io.read_jsonl(path_cached_genome))[0]
        if genome == cached_genome["genome"]:
            return True
    return False



from __future__ import annotations

import re
import pickle
import numpy as np
from pathlib import Path
from typing import Generator
from collections import defaultdict

from scipy import stats
from scipy.spatial import distance
from sklearn import linear_model

from DAJIN2.utils import io
from DAJIN2.core.preprocess import homopolymer_handler


def call_coverage_of_each_base(midsv_sample: Generator[dict], sequence: str) -> list[int]:
    coverages = [1] * len(sequence)
    for cont in midsv_sample:
        cssplits = cont["CSSPLIT"].split(",")
        for i, cssplit in enumerate(cssplits):
            if cssplit == "N":
                continue
            coverages[i] += 1
    return coverages


def count_indels(midsv_sample, sequence: str) -> dict[str, list[int]]:
    len_sequence = len(sequence)
    count = {"+": [0] * len_sequence, "-": [0] * len_sequence, "*": [0] * len_sequence}
    for samp in midsv_sample:
        for i, cs in enumerate(samp["CSSPLIT"].split(",")):
            if cs.startswith("=") or cs == "N" or re.search(r"a|c|g|t|n", cs):
                continue
            if cs.startswith("+"):
                # count["+"][i] += len(cs.split("|"))
                count["+"][i] += 1
            elif cs.startswith("-"):
                count["-"][i] += 1
            elif cs.startswith("*"):
                count["*"][i] += 1
    return count


def normalize_indels(count: dict[str, list[int]], coverages: list[int]) -> dict[str, np.array]:
    count_normalized = dict()
    coverages = np.array(coverages)
    for mut in count:
        counts = np.array(count[mut])
        count_normalized[mut] = counts / coverages
    return count_normalized


def split_kmer(indels: dict[str, np.array], kmer: int = 11) -> dict[str, np.array]:
    results = defaultdict(list)
    center = kmer // 2
    for mut, value in indels.items():
        for i in range(len(value)):
            if center <= i <= len(value) - center:
                start = i - center
                if kmer % 2 == 0:
                    end = i + center
                else:
                    end = i + center + 1
                results[mut].append(value[start:end])
            else:
                results[mut].append(np.array([0] * kmer))
    return results


def extract_dissimilar_loci(indels_kmer_sample: dict, indels_kmer_control: dict) -> dict[str, set]:
    """
    Comparing Sample and Control, the 'similar mean' and
    'similar variance' are considered as sequence errors.
    """
    results = dict()
    for mut in ["+", "-", "*"]:
        values_sample = indels_kmer_sample[mut]
        values_control = indels_kmer_control[mut]
        """Calculate cosine similarity: 1 means exactly same, 0 means completely different.
        - Zero vector does not return correct value, so add 1e-10.
            - example: distance.cosine([0,0,0], [1,2,3]) == 0
        """
        cossims = [1 - distance.cosine(x + 1e-10, y + 1e-10) for x, y in zip(values_sample, values_control)]
        # Perform T-test: nan means exactly same, p > 0.05 means similar in average.
        t_pvalues = [stats.ttest_ind(x, y, equal_var=False)[1] for x, y in zip(values_sample, values_control)]
        t_pvalues = [1 if np.isnan(t) else t for t in t_pvalues]
        # if pvalue == nan or pval > 0.05, samples and controls are similar.
        dissimilar_loci = set()
        for i, (cossim, t_pval) in enumerate(zip(cossims, t_pvalues)):
            if (cossim >= 0.8 and t_pval < 0.05) or cossim < 0.8:
                dissimilar_loci.add(i)
        results[mut] = dissimilar_loci
    return results


###########################################################
# Extract anomaly loci using OneClassSVM
# `extract_different_loci` does not consider the mutation rate in each kmer.
# Thus we got many false positives of kmer with the very low percentage of mutation rate
# Consider the mutation rate in the whole sequence
###########################################################


def _transform_log2(values: np.array) -> np.array:
    values = np.where(values <= 0, 1e-10, values)
    return np.log2(values).reshape(-1, 1)


def _merge_surrounding_index(idx_outliers: list) -> set:
    """If an outlier is found in an adjacent 5-mer, the area is also judged as an outlier."""
    idx_merged = set()
    for i, idx_curr in enumerate(idx_outliers):
        if i + 1 == len(idx_outliers):
            break
        idx_next = idx_outliers[i + 1]
        if idx_next - idx_curr <= 5:
            for j in range(idx_curr, idx_next + 1):
                idx_merged.add(j)
        else:
            idx_merged.add(idx_curr)
    return idx_merged


def extract_anomal_loci(indels_normalized_sample, indels_normalized_control) -> dict[str, set]:
    results = dict()
    for mut in ["+", "-", "*"]:
        # preprocess
        values_sample = indels_normalized_sample[mut]
        values_control = indels_normalized_control[mut]
        log2_subtract = _transform_log2(values_sample - values_control)
        # anomaly detection
        clf = linear_model.SGDOneClassSVM(random_state=0)
        predicts = clf.fit_predict(log2_subtract)
        # clf.fit(log2_control)
        # predicts = clf.predict(log2_sample)
        p1 = [i for i, p in enumerate(predicts) if p == -1]
        p2 = [i for i, p in enumerate(predicts) if p == 1]
        if np.mean(log2_subtract[p1]) > np.mean(log2_subtract[p2]):
            idx_outliers = p1
        else:
            idx_outliers = p2
        results[mut] = _merge_surrounding_index(idx_outliers)
    return results


###########################################################
# Homolopolymer region
###########################################################


def discard_errors_in_homopolymer(candidate_loci: dict[str, set], errors: dict[str, set]) -> dict[str, set]:
    mutation_loci = dict()
    for mut in ["+", "-", "*"]:
        mutation_loci[mut] = candidate_loci[mut] - errors[mut]
    return mutation_loci


###########################################################
# Merge non-contiguous insertions
###########################################################


def merge_index_of_consecutive_insertions(mutation_loci: dict[str, set[int]]) -> dict[str, set[int]]:
    """Treat as contiguous insertions if there are insertions within five bases of each other"""
    idx_ins_merged = set()
    idx_ins = sorted(mutation_loci["+"])
    for i in range(len(idx_ins) - 1):
        idx_1 = idx_ins[i]
        idx_2 = idx_ins[i + 1]
        if idx_1 + 5 > idx_2:
            for i in range(idx_1, idx_2 + 1):
                idx_ins_merged.add(i)
        else:
            idx_ins_merged.add(idx_1)
    mutation_loci["+"] = idx_ins_merged
    return mutation_loci


###########################################################
# main
###########################################################


def _process_control(TEMPDIR: Path, FASTA_ALLELES: dict, CONTROL_NAME: str) -> None:
    for allele, sequence in FASTA_ALLELES.items():
        path_mutation_loci = Path(TEMPDIR, CONTROL_NAME, "mutation_loci")
        if Path(path_mutation_loci, f"{allele}_count.pickle").exists():
            continue
        filepath_control = Path(TEMPDIR, CONTROL_NAME, "midsv", f"{allele}.json")
        indels_control = count_indels(io.read_jsonl(filepath_control), sequence)
        coverages_control = call_coverage_of_each_base(io.read_jsonl(filepath_control), sequence)
        indels_normalized_control = normalize_indels(indels_control, coverages_control)
        indels_kmer_control = split_kmer(indels_normalized_control, kmer=11)
        # Save indels_normalized_control and indels_kmer_control as pickle to reuse in consensus calling
        with open(Path(path_mutation_loci, f"{allele}_count.pickle"), "wb") as f:
            pickle.dump(indels_control, f)
        with open(Path(path_mutation_loci, f"{allele}_normalized.pickle"), "wb") as f:
            pickle.dump(indels_normalized_control, f)
        with open(Path(path_mutation_loci, f"{allele}_kmer.pickle"), "wb") as f:
            pickle.dump(indels_kmer_control, f)


def merge_loci(dissimilar_loci: dict[str, set], anomal_loci: dict[str, set]) -> dict[str, set]:
    mutation_loci = dict()
    for mut in ["+", "-", "*"]:
        mutation_loci[mut] = dissimilar_loci[mut] & anomal_loci[mut]
    return mutation_loci


def add_knockin_loci(candidate_loci: dict[str, set], knockin_loci: set):
    mutation_loci = dict()
    for mut in ["+", "-", "*"]:
        mutation_loci[mut] = candidate_loci[mut] | knockin_loci
    return mutation_loci


def transpose_mutation_loci(mutation_loci: set[int], sequence: str) -> list[set]:
    len_sequence = len(sequence)
    mutation_loci_transposed = [set() for _ in range(len_sequence)]
    for mut, idx_mutation in mutation_loci.items():
        for i, loci in enumerate(mutation_loci_transposed):
            if i in idx_mutation:
                loci.add(mut)
    return mutation_loci_transposed


def extract_mutation_loci(
    TEMPDIR: Path, FASTA_ALLELES: dict, SAMPLE_NAME: str, CONTROL_NAME: str, is_control=False
) -> None:
    if is_control:
        _process_control(TEMPDIR, FASTA_ALLELES, CONTROL_NAME)
        return

    for allele, sequence in FASTA_ALLELES.items():
        path_output = Path(TEMPDIR, SAMPLE_NAME, "mutation_loci", f"{allele}.pickle")
        if path_output.exists():
            continue

        filepath_sample = Path(TEMPDIR, SAMPLE_NAME, "midsv", f"{allele}.json")
        indels_sample = count_indels(io.read_jsonl(filepath_sample), sequence)
        coverages_sample = call_coverage_of_each_base(io.read_jsonl(filepath_sample), sequence)
        indels_normalized_sample = normalize_indels(indels_sample, coverages_sample)
        indels_kmer_sample = split_kmer(indels_normalized_sample, kmer=11)
        # Load indels_normalized_control and indels_kmer_control
        with open(Path(TEMPDIR, CONTROL_NAME, "mutation_loci", f"{allele}_normalized.pickle"), "rb") as f:
            indels_normalized_control = pickle.load(f)
        with open(Path(TEMPDIR, CONTROL_NAME, "mutation_loci", f"{allele}_kmer.pickle"), "rb") as f:
            indels_kmer_control = pickle.load(f)

        # Extract candidate mutation loci
        dissimilar_loci = extract_dissimilar_loci(indels_kmer_sample, indels_kmer_control)
        anomal_loci = extract_anomal_loci(indels_normalized_sample, indels_normalized_control)
        candidate_loci = merge_loci(dissimilar_loci, anomal_loci)

        # Extract error loci in homopolymer regions
        errors_in_homopolymer = homopolymer_handler.extract_errors(
            sequence, indels_normalized_sample, indels_normalized_control, candidate_loci
        )
        mutation_loci = discard_errors_in_homopolymer(candidate_loci, errors_in_homopolymer)

        # Merge all mutations and knockin loci
        path_knockin = Path(TEMPDIR, SAMPLE_NAME, "knockin_loci", f"{allele}.pickle")
        if path_knockin.exists():
            with open(path_knockin, "rb") as p:
                knockin_loci = pickle.load(p)
            mutation_loci = add_knockin_loci(mutation_loci, knockin_loci)

        mutation_loci = merge_index_of_consecutive_insertions(mutation_loci)
        mutation_loci_transposed = transpose_mutation_loci(mutation_loci, sequence)
        with open(path_output, "wb") as p:
            pickle.dump(mutation_loci_transposed, p)



from DAJIN2.core.preprocess import (
    fastx_parser,
    genome_fetcher,
    cache_checker,
    directories,
    mapping,
    midsv_caller,
)


from DAJIN2.core.preprocess.knockin_handler import extract_knockin_loci
from DAJIN2.core.preprocess.extract_mutation_loci import extract_mutation_loci
from DAJIN2.core.preprocess.generate_insertion_fasta import generate_insertion_fasta

# from DAJIN2.core.preprocess.extract_mutation_loci import process_mutation_loci
# from DAJIN2.core.preprocess.tmp_get_index_mapping import save_index_mapping



from __future__ import annotations

import pickle

from pathlib import Path
from typing import NamedTuple
from itertools import groupby
from collections import defaultdict

from DAJIN2.utils.cssplits_handler import find_n_boundaries

###########################################################
# call position weight matrix (cons_pergentage)
###########################################################


def _remove_nonconsecutive_n(cons_percentage: list[dict[str, float]]) -> list[dict[str, float]]:
    """remove nonconsecutive N"""
    cons_percentage_remove_n = []
    n_left, n_right = find_n_boundaries([max(c, key=c.get) for c in cons_percentage])
    for i, cons_per in enumerate(cons_percentage):
        if n_left < i < n_right:
            if not cons_per == {"N": 100}:
                cons_per.pop("N", None)
        cons_percentage_remove_n.append(cons_per)
    return cons_percentage_remove_n


def _update_percentage(cons_percentage: list[dict[str, float]]) -> list[dict[str, float]]:
    """replace sequence error as distributing according to proportion of cs tags"""
    cons_percentage_update = []
    for cons_per in cons_percentage:
        if "SEQERROR" not in cons_per:
            cons_percentage_update.append(cons_per)
            continue
        if len(cons_per) == 1 and cons_per["SEQERROR"]:
            cons_percentage_update.append({"N": 100})
            continue
        cons_per_update = dict()
        div = 100 / (sum(cons_per.values()) - cons_per["SEQERROR"])
        for key, val in cons_per.items():
            if key == "SEQERROR":
                continue
            cons_per_update[key] = val * div
        cons_percentage_update.append(cons_per_update)
    return cons_percentage_update


def _call_percentage(cssplits: list[str], mutation_loci) -> list[dict[str, float]]:
    """call position weight matrix in defferent loci.
    - non defferent loci are annotated to "Match" or "Unknown(N)"
    - sequence errors are annotated to "SEQERROR"
    """
    cssplits_transposed = [list(cs) for cs in zip(*cssplits)]
    coverage = len(cssplits)
    cons_percentage = []
    for cs_transposed, mut_loci in zip(cssplits_transposed, mutation_loci):
        count_cs = defaultdict(float)
        for cs in cs_transposed:
            if cs[0] in {"+", "-", "*"} and cs[0] not in mut_loci:
                cs = "SEQERROR"
            count_cs[cs] += 1 / coverage * 100
        cons_percentage.append(dict(count_cs))
    cons_percentage = _remove_nonconsecutive_n(cons_percentage)
    cons_percentage = _update_percentage(cons_percentage)
    return cons_percentage


###########################################################
# Call sequence
###########################################################


def _process_base(cons: str) -> str:
    if cons.startswith("="):  # match
        return cons.replace("=", "")
    elif cons.startswith("-"):  # deletion
        return ""
    elif cons.startswith("*"):  # substitution
        return cons[-1]
    elif cons.startswith("+"):  # insertion
        cons_ins = cons.split("|")
        if cons_ins[-1].startswith("="):  # match after insertion
            cons = cons.replace("=", "")
        elif cons_ins[-1].startswith("-"):  # deletion after insertion
            cons = "".join(cons_ins[:-1])
        elif cons_ins[-1].startswith("*"):  # substitution after insertion
            cons = "".join([*cons_ins[:-1], cons_ins[-1][-1]])
        return cons.replace("+", "").replace("|", "")
    return cons


def _call_sequence(cons_percentage: list[dict[str, float]]) -> str:
    consensus_sequence = []
    n_left, n_right = find_n_boundaries(cons_percentage)
    for i, cons_per in enumerate(cons_percentage):
        if n_left < i < n_right:
            cons = max(cons_per, key=cons_per.get)
            consensus_sequence.append(_process_base(cons))
        else:
            consensus_sequence.append("N")
    return "".join(consensus_sequence)


###########################################################
# main
###########################################################


class ConsensusKey(NamedTuple):
    allele: str
    label: int
    percent: float


def call_consensus(
    TEMPDIR: Path, SAMPLE_NAME: str, clust_sample: list[dict]
) -> tuple[defaultdict[list], defaultdict[str]]:
    cons_percentages = defaultdict(list)
    cons_sequences = defaultdict(str)
    clust_sample.sort(key=lambda x: x["LABEL"])
    for label, group in groupby(clust_sample, key=lambda x: x["LABEL"]):
        clust = list(group)
        allele = clust[0]["ALLELE"]
        key = ConsensusKey(allele, label, clust[0]["PERCENT"])
        with open(Path(TEMPDIR, SAMPLE_NAME, "mutation_loci", f"{allele}.pickle"), "rb") as p:
            mutation_loci = pickle.load(p)
        cssplits = [cs["CSSPLIT"].split(",") for cs in clust]
        cons_percentage = _call_percentage(cssplits, mutation_loci)
        cons_percentages[key] = cons_percentage
        cons_sequences[key] = _call_sequence(cons_percentage)
    return cons_percentages, cons_sequences



from __future__ import annotations

import midsv
from pathlib import Path
from itertools import groupby


def _calc_match(CSSPLIT: str) -> float:
    match_score = CSSPLIT.count("=")
    match_score -= CSSPLIT.count("+")  # insertion
    match_score -= sum(cs.islower() for cs in CSSPLIT)  # inversion
    cssplit = CSSPLIT.split(",")

    return match_score / len(cssplit)


def _score_allele(TEMPDIR: Path, allele: str, SAMPLE_NAME: str) -> list[dict]:
    midsv_sample = midsv.read_jsonl(Path(TEMPDIR, SAMPLE_NAME, "midsv", f"{allele}.json"))
    scored_alleles = []

    for dict_midsv in midsv_sample:
        score = _calc_match(dict_midsv["CSSPLIT"])
        dict_midsv.update({"SCORE": score, "ALLELE": allele})
        scored_alleles.append(dict_midsv)

    return scored_alleles


def _extract_alleles_with_max_score(score_of_each_alleles: list[dict]) -> list[dict]:
    alleles_with_max_score = []
    score_of_each_alleles.sort(key=lambda x: x["QNAME"])
    for _, group in groupby(score_of_each_alleles, key=lambda x: x["QNAME"]):
        max_read = max(group, key=lambda x: x["SCORE"])
        del max_read["SCORE"]
        alleles_with_max_score.append(max_read)
    return alleles_with_max_score


##########################################################
# main
##########################################################


def classify_alleles(TEMPDIR: Path, FASTA_ALLELES: dict, SAMPLE_NAME: str) -> list[dict]:
    score_of_each_alleles = []
    for allele in FASTA_ALLELES:
        score_of_each_alleles.extend(_score_allele(TEMPDIR, allele, SAMPLE_NAME))

    return _extract_alleles_with_max_score(score_of_each_alleles)



from __future__ import annotations

import re
from collections import defaultdict
import numpy as np
import scipy
from statsmodels.nonparametric.smoothers_lowess import lowess as sm_lowess


def get_repeat_regions(sequence: str, loci: set[int]) -> list[tuple[int, int]]:
    """
    Find homopolymers in the sequence but discard them that
    are adjacent to candidate mutation loci because they are
    likely to be covered by the real mutations
    """
    pattern = r"A{4,}|C{4,}|G{4,}|T{4,}|N{4,}"
    repeat_regions = []
    for start, end in (match.span() for match in re.finditer(pattern, sequence)):
        if not (start - 1 in loci and end + 1 in loci):
            repeat_regions.append((start, end))
    return repeat_regions


def get_counts_homopolymer(
    indels_sample_mut, indels_control_mut, repeat_regions
) -> tuple[dict[int, list[float]], list[tuple[int, int, np.array]]]:
    # Initialize default dictionaries to hold counts
    mutation_counts = defaultdict(list)
    mutation_counts_regions = []
    # Iterate through each repeat region
    for start, end in repeat_regions:
        # Calculate mutations for each sample and control
        sample_mutations = np.array(indels_sample_mut[start:end])
        control_mutations = np.array(indels_control_mut[start:end])
        # Total mutations is the sum of sample and control
        total_mutations = sample_mutations + control_mutations
        # If the total number of mutations is greater in the last position, reverse the order
        if total_mutations[0] > total_mutations[-1]:
            total_mutations = total_mutations[::-1]
            start, end = end, start
        # Apply a log transformation to the total mutations
        total_mutations_log = np.log(total_mutations)
        # Append the start, end, and total log mutations to the region count
        mutation_counts_regions.append((start, end, total_mutations_log))
        # Append the log mutation count to each position
        for position, value in enumerate(total_mutations_log):
            mutation_counts[position].append(value)
    return mutation_counts, mutation_counts_regions


def _smooth_data(input_x, input_y, input_xgrid):
    # Sample 50 data points from the input x and y
    sampled_indices = np.random.choice(len(input_x), 50, replace=True)
    sampled_y = input_y[sampled_indices]
    sampled_x = input_x[sampled_indices]
    # Apply lowess smoothing to the sampled data
    smoothed_y = sm_lowess(sampled_y, sampled_x, frac=1.0 / 5.0, it=5, return_sorted=False)
    # Interpolate the smoothed data onto the input grid
    interpolated_y_grid = scipy.interpolate.interp1d(sampled_x, smoothed_y, fill_value="extrapolate")(input_xgrid)
    return interpolated_y_grid


def return_thresholds(mutation_counts) -> list(float):
    # Initialize empty lists to hold x and y data
    mutation_positions = []
    mutation_counts_log = []
    # Populate the x and y data with the mutation counts and positions
    for position, counts in mutation_counts.items():
        position_values = [position] * len(counts)
        mutation_positions.extend(position_values)
        mutation_counts_log.extend(counts)
    # Convert x and y data to numpy arrays
    mutation_positions = np.array(mutation_positions)
    mutation_counts_log = np.array(mutation_counts_log)
    # Define a grid of x values from the minimum to the maximum position
    x_grid = np.linspace(mutation_positions.min(), mutation_positions.max(), mutation_positions.max() + 1)
    # Smooth the y data K times and stack the results
    num_smoothings = 100
    smoothed_data = np.stack(
        [_smooth_data(mutation_positions, mutation_counts_log, x_grid) for _ in range(num_smoothings)]
    ).T
    # Calculate the mean and standard error of the smoothed data
    mean_smoothed_data = np.nanmean(smoothed_data, axis=1)
    stderr_smoothed_data = scipy.stats.sem(smoothed_data, axis=1)
    stderr_smoothed_data = np.nanstd(smoothed_data, axis=1, ddof=0)
    # Define the thresholds
    thresholds = mean_smoothed_data + 1.90 * stderr_smoothed_data
    return thresholds


def get_errors_in_homopolyer(mutation_counts_regions, thresholds) -> set(int):
    # Initialize a set to hold the locations of mutations
    sequence_error_loci = set()
    # Iterate through each region and its associated mutation counts
    for start, end, log_mutations in mutation_counts_regions:
        # +-1 because 0-index is not considered as homopolymer
        if start > end:
            region = set(range(end, start - 1))
        else:
            region = set(range(start + 1, end))
        # Initialize a list to hold the indices of mutations
        mutation_indices = []
        # Iterate through each mutation count
        for index, log_mutation in enumerate(log_mutations):
            # Skip the first count (since it has no previous count to compare to)
            if index == 0:
                continue
            # If the count exceeds the threshold, record the index
            if log_mutation > thresholds[index]:
                mutation_indices.append(index)
        # Add the absolute location of each mutation to the set of mutation locations
        mutations = set()
        for mutation_index in mutation_indices:
            if start > end:
                mutations.add(start - 1 - mutation_index)
            else:
                mutations.add(end + mutation_index)
        sequence_error_loci |= region - mutations
    return sequence_error_loci


###########################################################
# main
###########################################################


def extract_errors(sequence, indels_sample, indels_control, candidate_loci: dict[set]) -> dict[str, set(int)]:
    errors_in_homopolymer = dict()
    for mut in ["+", "-", "*"]:
        repeat_regions = get_repeat_regions(sequence, candidate_loci[mut])
        if repeat_regions == []:
            errors_in_homopolymer[mut] = set()
            continue
        mutation_counts, mutation_counts_regions = get_counts_homopolymer(
            indels_sample[mut], indels_control[mut], repeat_regions
        )
        thresholds = return_thresholds(mutation_counts)
        errors_in_homopolymer[mut] = get_errors_in_homopolyer(mutation_counts_regions, thresholds)
    return errors_in_homopolymer



from __future__ import annotations

import re
from collections import defaultdict


def _detect_sv(cons_percentages: defaultdict[list], threshold: int = 50) -> list[bool]:
    exists_sv = []
    for cons_per in cons_percentages.values():
        cons_cssplits = []
        for cssplit in cons_per:
            seq = max(cssplit, key=cssplit.get)
            cons_cssplits.append(seq)
        cons_cssplits = "".join(cons_cssplits)
        if "N" * threshold in cons_cssplits:
            exists_sv.append(True)
        elif re.search(rf"(\+[ACGTN]\|){{{threshold}}}", cons_cssplits):
            exists_sv.append(True)
        elif re.search(rf"(\-[ACGTN]){{{threshold}}}", cons_cssplits):
            exists_sv.append(True)
        elif re.search(rf"(\*[ACGTN][ACGTN]){{{threshold}}}", cons_cssplits):
            exists_sv.append(True)
        elif re.search(r"[acgtn]", cons_cssplits):
            exists_sv.append(True)
        else:
            exists_sv.append(False)
    return exists_sv


def _format_allele_label(label: int, total_labels: int) -> str:
    label_digits = len(str(total_labels))
    return f"{label:0{label_digits}}"


def _determine_suffix(cons_seq: str, fasta_allele: str, is_sv: bool) -> str:
    if cons_seq == fasta_allele:
        return "_intact"
    elif is_sv:
        return "_sv"
    else:
        return "_indels"


def _construct_allele_name(
    label: int, allele: str, cons_seq: str, fasta_allele: str, percent: int, is_sv: bool, total_labels: int
) -> str:
    label_format = _format_allele_label(label, total_labels)
    suffix = _determine_suffix(cons_seq, fasta_allele, is_sv)
    return f"allele{label_format}_{allele}{suffix}_{percent}%"


def call_allele_name(
    cons_sequences: dict[int, dict],
    cons_percentages: dict[int, list],
    FASTA_ALLELES: dict[str, str],
    threshold: int = 50,
) -> dict[int, str]:
    exists_sv = _detect_sv(cons_percentages, threshold)
    total_labels = len(cons_percentages)
    allele_names = {}

    for is_sv, (keys, cons_seq) in zip(exists_sv, cons_sequences.items()):
        allele_name = _construct_allele_name(
            keys.label, keys.allele, cons_seq, FASTA_ALLELES[keys.allele], keys.percent, is_sv, total_labels
        )
        allele_names[keys.label] = allele_name

    return allele_names


def update_key_by_allele_name(cons: dict, allele_names: dict[int, str]) -> dict:
    cons_update = dict()
    for key in cons:
        old_allele = cons[key]
        new_allele = allele_names[key.label]
        cons_update[new_allele] = old_allele
    return cons_update


def add_key_by_allele_name(clust_sample: list[dict], allele_names: dict[int, str]) -> list[dict]:
    for clust in clust_sample:
        label = clust["LABEL"]
        clust["NAME"] = allele_names[label]
    return clust_sample



from __future__ import annotations

import cstag
import mappy

from pathlib import Path
from typing import Generator


def revcomp(sequence: str) -> str:
    """Return the reverse complement of the given DNA sequence."""
    complement = {"A": "T", "C": "G", "G": "C", "T": "A"}
    return "".join(complement[nt] for nt in sequence[::-1])


def to_sam(
    path_reference_fasta: Path, path_query_fastx: Path, preset: str = "map-ont", threads: int = 1, cslong: bool = True
) -> Generator[str]:
    """Align sequences using mappy and Convert PAF to SAM.

    Args:
        path_reference_fasta (Path): Path of reference fasta.
        path_query_fastx (Path): Path of query fasta/fastq.
        preset (str, optional): Alignment preset. Defaults to "map-ont".
        threads (int, optional): Number of threads to use. Defaults to 1.
        cslong (bool, optional): Use long formatted CS tag if True. Defaults to True.

    Yields:
        str: SAM formatted alignment.
    """
    path_reference_fasta = str(path_reference_fasta)
    path_query_fastx = str(path_query_fastx)

    SAM = [f"@SQ\tSN:{n}\tLN:{len(s)}" for n, s, _ in mappy.fastx_read(path_reference_fasta)]

    ref = mappy.Aligner(path_reference_fasta, preset=preset, n_threads=threads)
    if not ref:
        raise ValueError(f"Failed to load {path_reference_fasta}")

    for QUERY_NAME, QUERY_SEQ, QUERY_QUAL in mappy.fastx_read(path_query_fastx):
        for hit in ref.map(QUERY_SEQ, cs=True):
            query_seq = QUERY_SEQ.upper()
            query_qual = QUERY_QUAL

            # Report flag
            if hit.is_primary:
                flag = 0 if hit.strand == 1 else 16
            else:
                flag = 2048 if hit.strand == 1 else 2064

            # Handle reverse complement for negative strand
            if hit.strand == -1:
                query_seq = revcomp(query_seq)
                if query_qual:
                    query_qual = query_qual[::-1]

            # Append softclips to CIGAR
            cigar = hit.cigar_str
            if hit.q_st > 0:
                softclip = f"{hit.q_st}S"
                cigar = softclip + cigar if hit.strand == 1 else cigar + softclip
            if len(query_seq) - hit.q_en > 0:
                softclip = f"{len(query_seq) - hit.q_en}S"
                cigar = cigar + softclip if hit.strand == 1 else softclip + cigar

            # Convert to CS tag's long format
            cs = "cs:Z:" + hit.cs
            if cslong:
                cs = cstag.lengthen(hit.cs, cigar, query_seq)

            # Summarize
            alignment = [
                QUERY_NAME,
                str(flag),
                hit.ctg,
                str(hit.r_st + 1),
                str(hit.mapq),
                cigar,
                "*",
                "0",
                "0",
                query_seq,
                "*" if query_qual is None else query_qual,
                cs,
            ]

            SAM.append("\t".join(alignment))

    for record in SAM:
        yield record


def output_sam(
    TEMPDIR: Path,
    path_fasta: str | Path,
    name_fasta: str,
    path_fastq: str | Path,
    name_fastq: str,
    preset: str = "map-ont",
    threads: int = 1,
):
    sam = to_sam(path_fasta, path_fastq, preset=preset, threads=threads)
    output_sam = Path(TEMPDIR, name_fastq, "sam", f"{preset}_{name_fasta}.sam")
    output_sam.write_text("\n".join(sam))


########################################################################
# main
########################################################################


def generate_sam(temp_dir: Path, paths_fasta: list[str], path_fastq: str, name_fastq: str, threads: int) -> None:
    for path_fasta in paths_fasta:
        path_fasta = Path(path_fasta)
        output_sam(temp_dir, path_fasta, path_fasta.stem, path_fastq, name_fastq, preset="map-ont", threads=threads)
        output_sam(temp_dir, path_fasta, path_fasta.stem, path_fastq, name_fastq, preset="splice", threads=threads)


########################################################################
# Create faidx
########################################################################


def make_faidx(path_fasta: str | Path) -> str:
    fasta = Path(path_fasta).read_text().split()
    name, length, offset = fasta[0].strip(">"), len("".join(fasta[1:])), len(fasta[0]) + 1
    linebase, linewidth = len(fasta[1]), len(fasta[1]) + 1
    return "\t".join(map(str, [name, length, offset, linebase, linewidth])) + "\n"



from __future__ import annotations

from typing import Generator
import json
from pathlib import Path


def read_json(filepath: Path | str) -> Generator[dict[str, str]]:
    with open(filepath, "r") as f:
        for line in f:
            yield json.loads(line)


# def generate_mutation_kmers(path_sample: Path | str, mutation_loci: list[set], kmer: int = 5) -> Generator[list[str]]:
#     midsv_sample = read_json(path_sample)
#     kmer_n = ",".join(["N"] * kmer)
#     cssplits = (cs["CSSPLIT"].split(",") for cs in midsv_sample)
#     for cssplit in cssplits:
#         mutation_kmers = [kmer_n] * (kmer // 2)
#         for i in range((kmer // 2), len(cssplit) - (kmer // 2)):
#             if mutation_loci[i] == set():
#                 mutation_kmers.append(kmer_n)
#                 continue
#             mutations = [cssplit[j][0] for j in range(i - kmer + 1, i + 1)]
#             if "+" in mutations:
#                 for j in range(i - kmer + 1, i + 1):
#                     if cssplit[j][0] == "+":
#                         cssplit[j] = "+I" + cssplit[j].split("|")[-1]
#             if all(mutation in mutation_loci[j] for j, mutation in zip(range(i - kmer + 1, i + 1), mutations)):
#                 kmer = ",".join(cssplit[i - kmer + 1 : i + 1])
#                 mutation_kmers.append(kmer)
#             else:
#                 mutation_kmers.append(kmer_n)
#         mutation_kmers += [kmer_n] * (kmer // 2)
#         yield mutation_kmers


def generate_mutation_kmers(
    path_sample: Path | str, mutation_loci: list[set[str]], compress_ins: bool = True
) -> Generator[list[str]]:
    midsv_sample = read_json(path_sample)
    for cssplit in (cs["CSSPLIT"].split(",") for cs in midsv_sample):
        mutation_kmers = ["N,N,N"]
        for i in range(1, len(cssplit) - 1):
            if mutation_loci[i] == set():
                mutation_kmers.append("N,N,N")
                continue
            mutation = cssplit[i][0]  # +, - , *, =, N
            if mutation == "+":
                if compress_ins:
                    for j in range(i - 1, i + 2):
                        if cssplit[j].startswith("+") and not cssplit[j].startswith("+I"):
                            cssplit[j] = f"+I{cssplit[j].split('|')[-1]}"
                else:
                    for j in range(i - 1, i + 2):
                        if cssplit[j].startswith("+") and not cssplit[j][1].isdigit():
                            cssplit[j] = f"+{cssplit[j].count('|')}{cssplit[j].split('|')[-1]}"
            if mutation in mutation_loci[i]:
                kmer = ",".join([cssplit[i - 1], cssplit[i], cssplit[i + 1]])
                mutation_kmers.append(kmer)
            else:
                mutation_kmers.append("N,N,N")
        mutation_kmers.append("N,N,N")
        yield mutation_kmers



from __future__ import annotations


def find_n_boundaries(cssplits: list[str]) -> tuple[int, int]:
    """Find the boundaries of contiguous Ns which aren't at the ends."""

    # Find the left boundary
    left_idx_n = 0
    for char in cssplits:
        if char != "N":
            break
        left_idx_n += 1

    # Find the right boundary
    right_idx_n = len(cssplits) - 1
    for char in reversed(cssplits):
        if char != "N":
            break
        right_idx_n -= 1

    return left_idx_n - 1, right_idx_n + 1



from __future__ import annotations

import re
import midsv
from pathlib import Path
from itertools import groupby
from DAJIN2.core import preprocess

###########################################################
# reverse complement to cssplits
###########################################################


def _reverse_cssplits(cssplits: list) -> list:
    for i, cs in enumerate(cssplits):
        if cs.startswith("+"):
            cssplits[i] = "+" + "|".join(cs.split("|")[::-1])
    return cssplits[::-1]


def _realign_insertion(cssplits: list) -> list:
    for i, cs in enumerate(cssplits):
        if not cs.startswith("+"):
            continue
        if re.search(rf"[{cs[1]}]", "[ACGTacgt]"):
            continue
        if i + 1 == len(cssplits):
            continue
        cs_current = cs.split("|")
        cssplits[i] = cs_current[0].replace("+", "")
        cssplits[i + 1] = "|".join(c[0] + c[-1] for c in cs_current[1:]) + "|" + cssplits[i + 1]
    return cssplits


def _complement_cssplit(cssplits: list) -> list:
    comp = {"A": "T", "C": "G", "G": "C", "T": "A", "N": "N", "a": "t", "c": "g", "g": "c", "t": "a", "n": "n"}
    for i, cs in enumerate(cssplits):
        op = cs[0]
        if op == "*":
            cssplits[i] = op + comp[cs[1]] + comp[cs[2]]
        elif op == "+":
            cssplits[i] = "|".join(c[0] + comp[c[-1]] for c in cs.split("|"))
        else:  # Match or Deletion or N
            cssplits[i] = op + comp[cs[-1]]
        cssplits[i] = cssplits[i].replace("NN", "N")
        cssplits[i] = cssplits[i].replace("nn", "n")
    return cssplits


def revcomp_cssplits(cssplits: list[str]) -> list[str]:
    cssplits_reversed = _reverse_cssplits(cssplits)
    cssplits_realigned = _realign_insertion(cssplits_reversed)
    cssplits_revcomped = _complement_cssplit(cssplits_realigned)
    return cssplits_revcomped


###########################################################
# group by mutation
###########################################################


def annotate_inversion(cssplits: list[str]) -> list[str]:
    return ["@" + cs if cs[-1].islower() else cs for cs in cssplits]


def group_by_mutation(cssplits: list[str]) -> list[list[str]]:
    return [list(group) for _, group in groupby(cssplits, key=lambda x: x[0])]


def flatten(lst: list[list]) -> list:
    results_flattend = []
    for result in lst:
        if isinstance(result[0], list):
            for res in result:
                results_flattend.append(res)
        else:
            results_flattend.append(result)
    return results_flattend


###########################################################
# report mutations
###########################################################


def _handle_match(group, genome, start, end, header, chromosome):
    end += len(group)
    start = end
    return None, start, end


def _handle_substitution(group, genome, start, end, header, chromosome):
    result = []
    for g in group:
        ref = g[1]
        mut = g[2]
        result.append([header, genome, chromosome, start, end, f"substitution: {ref}>{mut}"])
        end += 1
        start = end
    if len(result) == 1:
        result = result[0]
    return result, start, end


def _handle_deletion(group, genome, start, end, header, chromosome):
    end += len(group) - 1
    size = len(group)
    seq = "".join([g[-1] for g in group])
    result = [header, genome, chromosome, start, end, f"{size}bp deletion: {seq}"]
    end += 1
    start = end
    return result, start, end


def _handle_insertion(group, genome, start, end, header, chromosome):
    group = group[0]
    size = group.count("|")
    seq_insertion = "".join([g[-1] for g in group.split("|")[:-1]])
    seq_last = group.split("|")[-1]
    result = []
    result.append([header, genome, chromosome, start, end, f"{size}bp insertion: {seq_insertion}"])
    if seq_last.startswith("="):
        pass
    elif seq_last.startswith("-"):
        result.append([header, genome, chromosome, start, end, f"1bp deletion: {seq_last[-1]}"])
    if len(result) == 1:
        result = result[0]
    end += 1
    start = end
    return result, start, end


def _handle_inversion(group, genome, start, end, header, chromosome):
    end += len(group) - 1
    size = len(group)
    seq = "".join([g[-1].upper() for g in group])
    result = [header, genome, chromosome, start, end, f"{size}bp inversion: {seq}"]
    end += 1
    start = end
    return result, start, end


def _handle_unknown(group, genome, start, end, header, chromosome):
    end += len(group) - 1
    size = len(group)
    result = [header, genome, chromosome, start, end, f"{size}bp unknown bases"]
    end += 1
    start = end
    return result, start, end


def report_mutations(cssplits_grouped, GENOME_COODINATES, header):
    genome = GENOME_COODINATES["genome"]
    chromosome = GENOME_COODINATES["chrom"]
    start = end = GENOME_COODINATES["start"]
    handlers = {
        "=": _handle_match,
        "*": _handle_substitution,
        "-": _handle_deletion,
        "+": _handle_insertion,
        "@": _handle_inversion,
        "N": _handle_unknown,
    }
    results = []
    for group in cssplits_grouped:
        for prefix, handler in handlers.items():
            if group[0].startswith(prefix):
                result, start, end = handler(group, genome, start, end, header, chromosome)
                if prefix == "=":
                    continue
                results.append(result)
    return flatten(results)


###########################################################
# main
###########################################################


def to_csv(TEMPDIR: Path | str, SAMPLE_NAME: str, GENOME_COODINATES: dict, cons_percentage: dict) -> None:
    results = [["Allele ID", "Genome", "Chromosome", "Start", "End", "Mutation"]]
    ref = Path(TEMPDIR, SAMPLE_NAME, "fasta", "control.fasta")
    for query in Path(TEMPDIR, "report", "FASTA", SAMPLE_NAME).iterdir():
        sam = preprocess.mapping.to_sam(ref, query)
        sam = [s.split("\t") for s in sam]
        midsv_sample = midsv.transform(sam, midsv=False, cssplit=True, qscore=False)[0]
        header = midsv_sample["QNAME"]
        cssplits = midsv_sample["CSSPLIT"].split(",")
        # for header, cons_per in cons_percentage.items():
        #     cssplits = [max(cons, key=cons.get) for cons in cons_per]
        if GENOME_COODINATES["strand"] == "-":
            cssplits = revcomp_cssplits(cssplits)
        cssplits_inversion = annotate_inversion(cssplits)
        cssplits_grouped = group_by_mutation(cssplits_inversion)
        result = report_mutations(cssplits_grouped, GENOME_COODINATES, header)
        results.extend(result)
    results_csv = "\n".join([",".join(map(str, r)) for r in results]) + "\n"
    path_output = Path(TEMPDIR, "report", "MUTATION_INFO", f"{SAMPLE_NAME}.csv")
    path_output.write_text(results_csv)



from __future__ import annotations

import pickle
from collections import defaultdict
from itertools import groupby
from pathlib import Path

import midsv
import numpy as np
from rapidfuzz import process
from rapidfuzz.distance import DamerauLevenshtein
from sklearn import metrics
from sklearn.cluster import MeanShift, MiniBatchKMeans

###########################################################
# Detect insertion sequences
###########################################################


def _count_insertions(path: Path | str, mutation_loci: dict) -> dict[tuple[int, str], int]:
    insertion_counts = defaultdict(int)
    for m in midsv.read_jsonl(path):
        cssplits = m["CSSPLIT"].split(",")
        for idx in (i for i, m in enumerate(mutation_loci) if "+" in m):
            if cssplits[idx].startswith("+"):
                insertion_counts[(idx, cssplits[idx])] += 1
    # remove low frequency insertions
    coverage_sample = sum(1 for _ in midsv.read_jsonl(path))
    threshold = int(coverage_sample * 0.5 / 100)
    insertion_counts = {k: v for k, v in insertion_counts.items() if v >= threshold}
    return insertion_counts


def _create_insertions_dict(
    insertion_sample: dict[tuple[int, str], int], insertion_control: dict[tuple[int, str], int]
) -> defaultdict[dict[str, int]]:
    insertions = defaultdict(dict)
    for key in insertion_sample.keys() - insertion_control.keys():
        score = insertion_sample[key]
        idx, seq = key
        if isinstance(seq, int):
            idx, seq = seq, idx
        insertions[idx].update({seq: score})
    return insertions


def extract_insertions(
    path_sample: Path | str, path_control: Path | str, mutation_loci: dict
) -> defaultdict[dict[str, int]]:
    insertion_sample = _count_insertions(path_sample, mutation_loci)
    insertion_control = _count_insertions(path_control, mutation_loci)
    return _create_insertions_dict(insertion_sample, insertion_control)


###########################################################
# Merge similar insertion sequences
###########################################################


def _group_index_by_consecutive_insertions(mutation_loci: dict[str, set[int]]) -> list[tuple[int]]:
    index = sorted(i for i, m in enumerate(mutation_loci) if "+" in m)
    index_grouped = []
    for _, group in groupby(enumerate(index), lambda i_x: i_x[0] - i_x[1]):
        items = [value for _, value in group]
        index_grouped.append(tuple(items))
    return index_grouped


def _group_insertions(insertions, index_grouped) -> dict[dict[str, int]]:
    """
    Insertion bases in consecutive insertion base positions are grouped together in mutation_loci,
    as the insertion site may shift by a few bases.
    """
    insertions_grouped = defaultdict(dict)
    for idx in index_grouped:
        for i in idx:
            if i not in insertions:
                continue
            insertions_grouped[idx].update(insertions[i])
    return insertions_grouped


def _get_normalized_scores(query: str, choices: list[str]) -> np.ndarray:
    seqs, scores, _ = zip(*process.extract_iter(query, choices, scorer=DamerauLevenshtein.normalized_distance))
    scores_np = np.array(scores).reshape(-1, 1)
    normalized_scores = (scores_np - scores_np.min()) / (scores_np.max() - scores_np.min())
    counts = np.array([s.count("|") for s in seqs]).reshape(-1, 1)
    X = np.concatenate([normalized_scores, counts], axis=1)
    return X


def _optimize_labels(X: np.array) -> list[int]:
    sample_size = X.shape[0]
    labels_prev = list(range(sample_size))
    for i in range(1, sample_size):
        np.random.seed(seed=1)
        labels_current = MiniBatchKMeans(n_clusters=i, random_state=1, n_init="auto").fit_predict(X).tolist()
        silhuette = metrics.silhouette_score(X, labels_current) if i > 1 else 0
        mutual_info = metrics.adjusted_mutual_info_score(labels_prev, labels_current)
        # print(i, Counter(labels_current), round(silhuette, 2), round(mutual_info, 2) ) # ! DEBUG
        if i == 2 and silhuette < 0:
            return [1] * len(labels_current)
        if 0.9 < silhuette or 0.9 < mutual_info:
            return labels_current
        labels_prev = labels_current
    return labels_current


def _get_merged_insertion(insertion: dict[str, int], labels: np.ndarray) -> dict[frozenset, int]:
    insertion_label = [(label, {key: val}) for label, (key, val) in zip(labels, insertion.items())]
    insertion_label.sort(key=lambda x: x[0])
    insertion_merged = dict()
    for _, group in groupby(insertion_label, key=lambda x: x[0]):
        group = [g[1] for g in group]
        seq, count = set(), 0
        for g in group:
            key, val = list(g.items())[0]
            seq.add(key)
            count += val
        insertion_merged[frozenset(seq)] = count
    return insertion_merged


def merge_similar_insertions(insertions, mutation_loci) -> dict[dict[frozenset[str], int]]:
    index_grouped = _group_index_by_consecutive_insertions(mutation_loci)
    insertions_grouped = _group_insertions(insertions, index_grouped)
    insertions_merged = defaultdict(dict)
    for idx, insertion in insertions_grouped.items():
        if len(insertion) == 1:
            key, val = list(insertion.items())[0]
            insertions_merged[idx][frozenset([key])] = val
            continue
        query = list(insertion.keys())[0]
        X = _get_normalized_scores(query, list(insertion.keys()))
        labels = _optimize_labels(X)
        insertions_merged[idx] = _get_merged_insertion(insertion, labels)
    return insertions_merged


###########################################################
# Cluster insertion alleles
###########################################################


def extract_score_and_sequence(path_sample, insertions_merged) -> list[tuple[list[int], str]]:
    scores = []
    sequences = []
    for m in midsv.read_jsonl(path_sample):
        score = defaultdict(int)
        seq = defaultdict(lambda: "N")
        cssplits = m["CSSPLIT"].split(",")
        for idx_grouped, ins in insertions_merged.items():
            score[idx_grouped] = 0
            seq[idx_grouped] = "N"
            for idx in idx_grouped:
                for seqs, value in ins.items():
                    if cssplits[idx] in seqs:
                        score[idx_grouped] = value
                        seq[idx_grouped] = cssplits[idx]
        if any(score.values()):
            scores.append([s for _, s in sorted(score.items())])
            sequences.append(",".join(s for _, s in sorted(seq.items())))
    return [(score, sequence) for score, sequence in zip(scores, sequences)]


def clustering_insertions(scores) -> list[int]:
    X = np.array(scores)
    if X.max() - X.min() == 0:
        return [1] * len(X)
    X = (X - X.min()) / (X.max() - X.min())
    clustering = MeanShift().fit(X)
    labels = clustering.labels_
    return labels


###########################################################
# Call consensus
###########################################################


def subset_sequences(sequences, labels, num=1000) -> list[dict]:
    """
    Downsampling to the required number of sequences when there is a huge number of inserted sequences
    """
    sequences_subset = []
    tmp_sequences = []
    for sequence, label in zip(sequences, labels):
        tmp_sequences.append({"CSSPLIT": sequence, "LABEL": label})
    tmp_sequences.sort(key=lambda x: x["LABEL"])
    for _, group in groupby(tmp_sequences, key=lambda x: x["LABEL"]):
        sequences_subset.extend(list(group)[:num])
    return sequences_subset


def _call_percentage(cssplits: list[str]) -> list[dict[str, float]]:
    """call position weight matrix in defferent loci."""
    coverage = len(cssplits)
    cssplits_transposed = [list(cs) for cs in zip(*cssplits)]
    cons_percentage = []
    for cs_transposed in cssplits_transposed:
        count_cs = defaultdict(float)
        for cs in cs_transposed:
            count_cs[cs] += 1 / coverage * 100
        cons_percentage.append(count_cs)
    return cons_percentage


def _call_sequence(cons_percentage: list[dict[str, float]]) -> str:
    consensus_sequence = []
    for cons_per in cons_percentage:
        cons = max(cons_per, key=cons_per.get)
        if cons.startswith("="):
            cons = cons.replace("=", "")
        elif cons.startswith("-"):
            continue
        elif cons.startswith("*"):
            cons = cons[-1]
        elif cons.startswith("+"):
            cons_ins = cons.split("|")
            if cons_ins[-1].startswith("="):
                cons = cons.replace("=", "")
            elif cons_ins[-1].startswith("-"):
                cons = "".join(cons_ins[:-1])
            elif cons_ins[-1].startswith("*"):
                cons = "".join([*cons_ins[:-1], cons_ins[-1][-1]])
            cons = cons.replace("+", "")
            cons = cons.replace("|", "")
        consensus_sequence.append(cons)
    return ",".join(consensus_sequence)


def _remove_all_n(cons_sequence: dict[int, str]) -> dict[int, str]:
    cons_sequence_removed = dict()
    for label, seq in cons_sequence.items():
        if all(True if s == "N" else False for s in seq.split(",")):
            continue
        cons_sequence_removed[label] = seq
    return cons_sequence_removed


def call_consensus(sequences_subset: list[dict]) -> dict[int, str]:
    cons_sequence = dict()
    sequences_subset.sort(key=lambda x: x["LABEL"])
    for label, group in groupby(sequences_subset, key=lambda x: x["LABEL"]):
        cssplits = [cs["CSSPLIT"].split(",") for cs in group]
        cons_per = _call_percentage(cssplits)
        cons_seq = _call_sequence(cons_per)
        cons_sequence[label] = cons_seq
    return _remove_all_n(cons_sequence)


def extract_index_of_insertions(insertions, insertions_merged) -> list[int]:
    index_of_insertions = []
    for idx_group in insertions_merged:
        max_val = -1
        for idx in idx_group:
            if max_val < sum(insertions[idx].values()):
                max_val = sum(insertions[idx].values())
                max_idx = idx
        index_of_insertions.append(max_idx)
    return index_of_insertions


def generate_consensus(cons_sequence, index_of_insertions, sequence) -> dict[int, str]:
    consensus_sequence_insertion = dict()
    for label, cons_seq in cons_sequence.items():
        cons_seq = cons_seq.split(",")
        list_sequence = list(sequence)
        for idx, seq in zip(index_of_insertions, cons_seq):
            if seq == "N":
                continue
            list_sequence[idx] = seq
        consensus_sequence_insertion[label] = "".join(list_sequence)
    return consensus_sequence_insertion


def filter_consensus(consensus_sequence_insertion: dict[int, str], FASTA_ALLELES: dict) -> dict[int, str]:
    """Filter similar insertions compared to control sequence"""
    unique_insertions = set(consensus_sequence_insertion.values())
    for query in FASTA_ALLELES.values():
        seqs, mismatches, _ = zip(*process.extract_iter(query, unique_insertions, scorer=DamerauLevenshtein.distance))
        for seq, mismatch in zip(seqs, mismatches):
            if mismatch <= 10:
                unique_insertions.remove(seq)
    return {i: seq for i, seq in enumerate(unique_insertions)}


###########################################################
# generate fasta
###########################################################


def update_labels(d: dict, FASTA_ALLELES: dict) -> dict:
    user_defined_alleles = set(FASTA_ALLELES)
    d_values = list(d.values())
    len_d = len(d_values)
    digits_up = 0
    # Update labels to avoid duplicating user-specified alleles
    # (insertion1 -> insertion01 -> insertion001...)
    while True:
        digits = len(str(len_d)) + digits_up
        d_updated = {f"insertion{i+1:0{digits}}": seq for i, seq in enumerate(d_values)}
        if user_defined_alleles.isdisjoint(set(d_updated)):
            break
        digits_up += 1
    return d_updated


def save_fasta(TEMPDIR: Path | str, SAMPLE_NAME: str, consensus_sequence_insertion: dict) -> None:
    for header, seq in consensus_sequence_insertion.items():
        Path(TEMPDIR, SAMPLE_NAME, "fasta", f"{header}.fasta").write_text(f">{header}\n{seq}")


###########################################################
# main
###########################################################


def generate_insertion_fasta(TEMPDIR, SAMPLE_NAME, CONTROL_NAME, FASTA_ALLELES) -> None:
    path_sample = Path(TEMPDIR, SAMPLE_NAME, "midsv", "control.json")
    path_control = Path(TEMPDIR, CONTROL_NAME, "midsv", "control.json")
    with open(Path(TEMPDIR, SAMPLE_NAME, "mutation_loci", "control.pickle"), "rb") as p:
        mutation_loci = pickle.load(p)
    sequence = FASTA_ALLELES["control"]
    insertions = extract_insertions(path_sample, path_control, mutation_loci)
    index_set = set(i for i, m in enumerate(mutation_loci) if "+" in m)
    if insertions.keys() & index_set == set():
        return None
    insertions_merged = merge_similar_insertions(insertions, mutation_loci)
    insertions_scores_sequences = extract_score_and_sequence(path_sample, insertions_merged)
    labels = clustering_insertions([score for score, _ in insertions_scores_sequences])
    insertion_sequences_subset = subset_sequences([seq for _, seq in insertions_scores_sequences], labels, num=1000)
    cons_sequence = call_consensus(insertion_sequences_subset)
    index_of_insertions = extract_index_of_insertions(insertions, insertions_merged)
    consensus_sequence_insertion = generate_consensus(cons_sequence, index_of_insertions, sequence)
    if len(consensus_sequence_insertion) == 1:
        consensus_filtered = consensus_sequence_insertion
    else:
        consensus_filtered = filter_consensus(consensus_sequence_insertion, FASTA_ALLELES)
        if consensus_filtered == dict():
            return None
    consensus_labeled = update_labels(consensus_filtered, FASTA_ALLELES)
    save_fasta(TEMPDIR, SAMPLE_NAME, consensus_labeled)



from __future__ import annotations

import re
from pathlib import Path

import mappy

########################################################################
# Helper function
########################################################################


def _sanitize_name(name: str) -> str:
    """
    Sanitize the name by replacing invalid characters with '-'
    """
    name = name.lstrip()
    if not name:
        raise ValueError("Provided FASTA/FASTQ is empty or consists only of whitespace")
    return re.sub(r'[\\/:?.,\'"<>| ]', "-", name)


########################################################################
# Extract basename
########################################################################


def extract_basename(fastq_path: str) -> str:
    name = Path(fastq_path).name
    name = re.sub(r"\..*$", "", name)  # Remove file extension
    return _sanitize_name(name)


########################################################################
# Convert allele file to dictionary type fasta format
########################################################################


def dictionize_allele(path_fasta: str) -> dict:
    return {_sanitize_name(name): seq.upper() for name, seq, _ in mappy.fastx_read(path_fasta)}


########################################################################
# Export fasta files as single-FASTA format
########################################################################


def export_fasta_files(TEMPDIR: Path, FASTA_ALLELES: dict, NAME: str) -> None:
    """
    This function exports FASTA files in single-FASTA format.

    :param TEMPDIR: Temporary directory Path object where the output files will be saved.
    :param FASTA_ALLELES: Dictionary containing identifier and sequence pairs.
    :param NAME: Name to be included in the output path.
    """
    for identifier, sequence in FASTA_ALLELES.items():
        contents = "\n".join([">" + identifier, sequence]) + "\n"
        output_fasta = Path(TEMPDIR, NAME, "fasta", f"{identifier}.fasta")
        output_fasta.write_text(contents)



from __future__ import annotations

import re
import json
import pickle
import numpy as np
from pathlib import Path
from typing import Generator
from collections import defaultdict
from scipy import stats
from scipy.spatial import distance
from sklearn import linear_model


def read_midsv(filepath: str | Path) -> Generator[dict[str, str]]:
    with open(filepath, "r") as f:
        for line in f:
            yield json.loads(line)


def call_coverage_on_each_base(midsv_sample: Generator[dict], sequence: str) -> list[int]:
    coverages = [1] * len(sequence)
    for cont in midsv_sample:
        cssplits = cont["CSSPLIT"].split(",")
        for i, cssplit in enumerate(cssplits):
            if cssplit == "N":
                continue
            coverages[i] += 1
    return coverages


def count_indels(midsv_sample, sequence: str) -> dict[str, list[int]]:
    len_sequence = len(sequence)
    count = {"+": [0] * len_sequence, "-": [0] * len_sequence, "*": [0] * len_sequence}
    for samp in midsv_sample:
        for i, cs in enumerate(samp["CSSPLIT"].split(",")):
            if cs.startswith("=") or cs == "N" or re.search(r"a|c|g|t|n", cs):
                continue
            if cs.startswith("+"):
                # count["+"][i] += len(cs.split("|"))
                count["+"][i] += 1
            elif cs.startswith("-"):
                count["-"][i] += 1
            elif cs.startswith("*"):
                count["*"][i] += 1
    return count


def normalize_indels(count: dict[str, list[int]], coverages: list[int]) -> dict[str, np.array]:
    count_normalized = dict()
    coverages = np.array(coverages)
    for mut in count:
        counts = np.array(count[mut])
        count_normalized[mut] = counts / coverages
    return count_normalized


def split_kmer(indels: dict[str, np.array], kmer: int = 11) -> dict[str, np.array]:
    results = defaultdict(list)
    center = kmer // 2
    for mut, value in indels.items():
        for i in range(len(value)):
            if center <= i <= len(value) - center:
                start = i - center
                if kmer % 2 == 0:
                    end = i + center
                else:
                    end = i + center + 1
                results[mut].append(value[start:end])
            else:
                results[mut].append(np.array([0] * kmer))
    return results


def extract_dissimilar_loci(indels_kmer_sample: dict, indels_kmer_control: dict) -> dict[str, set]:
    """
    Comparing Sample and Control, the 'similar mean' and
    'similar variance' are considered as sequence errors.
    """
    results = dict()
    for mut in ["+", "-", "*"]:
        values_sample = indels_kmer_sample[mut]
        values_control = indels_kmer_control[mut]
        """Calculate cosine similarity: 1 means exactly same, 0 means completely different.
        - Zero vector does not return correct value, so add 1e-10.
            - example: distance.cosine([0,0,0], [1,2,3]) == 0
        """
        cossims = [1 - distance.cosine(x + 1e-10, y + 1e-10) for x, y in zip(values_sample, values_control)]
        # Perform T-test: nan means exactly same, p > 0.05 means similar in average.
        t_pvalues = [stats.ttest_ind(x, y, equal_var=False)[1] for x, y in zip(values_sample, values_control)]
        t_pvalues = [1 if np.isnan(t) else t for t in t_pvalues]
        # if pvalue == nan or pval > 0.05, samples and controls are similar.
        dissimilar_loci = set()
        for i, (cossim, t_pval) in enumerate(zip(cossims, t_pvalues)):
            if (cossim >= 0.8 and t_pval < 0.05) or cossim < 0.8:
                dissimilar_loci.add(i)
        results[mut] = dissimilar_loci
    return results

###########################################################
# Extract anomaly loci using OneClassSVM
# `extract_different_loci` does not consider the mutation rate in each kmer.
# Thus we got many false positive of kmer with the low percentage of mutation rate
# Consider the mutation rate in the sequence
###########################################################


def _transform_log2(values: np.array) -> np.array:
    values = np.where(values <= 0, 1e-10, values)
    return np.log2(values).reshape(-1, 1)


def _get_divisor(set1, set2) -> int:
    return len(set(set1) & set(set2)) or -1


def _merge_peaks(log2_sample, log2_control, peaks) -> set:
    """Values higher than 75% quantile of the control values and
    the surrouings are peaks, so merge it as a peak"""
    threshold = np.quantile(log2_control, 0.75)
    for i, value in enumerate(log2_sample):
        if i not in peaks and value > threshold:
            for j in range(i - 5, i + 6):
                if j in peaks:
                    peaks.add(i)
                    break
    return peaks


def extract_anomal_loci(indels_sample_normalized, indels_control_normalized) -> dict[str, set]:
    results = dict()
    for mut in ["+", "-", "*"]:
        # preprocess
        values_sample = indels_sample_normalized[mut]
        values_control = indels_control_normalized[mut]
        values_subtract = _transform_log2(values_sample - values_control)
        log2_control = _transform_log2(values_control)
        log2_sample = _transform_log2(values_sample)
        # anomaly detection
        clf = linear_model.SGDOneClassSVM(random_state=0)
        clf.fit(log2_control)
        predicts = clf.predict(log2_sample)
        idx_outliers = np.where(predicts == -1)[0]
        idx_outliers_reverse = np.where(predicts == 1)[0]
        # anomaly detection by quantile
        idx_upper = np.where(values_subtract > np.quantile(log2_control, 0.75))[0]
        # determine which is the correct outliers that is the percentage of outliers is large
        idx1 = len(idx_outliers) / _get_divisor(idx_outliers, idx_upper)
        idx2 = len(idx_outliers_reverse) / _get_divisor(idx_outliers_reverse, idx_upper)
        if idx1 < idx2:
            idx_outliers = idx_outliers_reverse
        results[mut] = _merge_peaks(log2_sample, log2_control, set(idx_outliers) & set(idx_upper))
    return results


###########################################################
# Homolopolymer region
###########################################################


def discard_errors_in_homopolymer(dissimilar_loci: dict[str, set], errors_in_homopolymer: dict[str, set]) -> dict[str, set]:
    mutation_loci = dict()
    for mut in ["+", "-", "*"]:
        error_loci = errors_in_homopolymer[mut]
        mutation_loci[mut] = dissimilar_loci[mut] - error_loci
    return mutation_loci


###########################################################
# Biased strand
###########################################################

def discard_errors_on_biased_strand(midsv_sample, mutation_loci) -> dict[str, set]:
    results = dict()
    for mutation, loci in mutation_loci.items():
        mutation_loci_non_biased = set()
        count_plus = defaultdict(int)
        count_total = defaultdict(int)
        midsv_sample = list(midsv_sample)
        for samp in midsv_sample:
            cssplits = samp["CSSPLIT"].split(",")
            for i, cs in enumerate(cssplits):
                if i not in loci:
                    continue
                if not cs.startswith(mutation):
                    continue
                count_total[i] += 1
                if samp["STRAND"] == "+":
                    count_plus[i] += 1
        mutation_loci_non_biased = {i for i, total in count_total.items() if 0.25 < (count_plus[i] / total) < 0.75}
        results[mutation] = mutation_loci_non_biased
    return results

###########################################################
# main
###########################################################


def process_mutation_loci(TEMPDIR: Path, FASTA_ALLELES: dict, CONTROL_NAME: str) -> None:
    for allele, sequence in FASTA_ALLELES.items():
        filepath_control = Path(TEMPDIR, "midsv", f"{CONTROL_NAME}_{allele}.json")
        indels_control = count_indels(read_midsv(filepath_control), sequence)
        coverages_control = call_coverage_on_each_base(read_midsv(filepath_control), sequence)
        indels_control_normalized = normalize_indels(indels_control, coverages_control)
        indels_kmer_control = split_kmer(indels_control_normalized, kmer=11)
        # Save indels_control_normalized and indels_kmer_control as pickle to reuse in consensus calling
        with open(Path(TEMPDIR, "mutation_loci", f"{CONTROL_NAME}_{allele}..pickle"), "wb") as f:
            pickle.dump(indels_control, f)
        with open(Path(TEMPDIR, "mutation_loci", f"{CONTROL_NAME}_{allele}_normalized..pickle"), "wb") as f:
            pickle.dump(indels_control_normalized, f)
        with open(Path(TEMPDIR, "mutation_loci", f"{CONTROL_NAME}_{allele}_kmer..pickle"), "wb") as f:
            pickle.dump(indels_kmer_control, f)


def is_strand_bias(midsv_control) -> bool:
    count_strand = defaultdict(int)
    for m in midsv_control:
        count_strand[m["STRAND"]] += 1
    percentage_plus = count_strand["+"] / (count_strand["+"] + count_strand["-"])
    if 0.25 < percentage_plus < 0.75:
        return False
    else:
        return True


def merge_loci(dissimilar_loci: dict[str, set], anomal_loci: dict[str, set]) -> dict[str, set]:
    mutation_loci = dict()
    for mut in ["+", "-", "*"]:
        mutation_loci[mut] = dissimilar_loci[mut] & anomal_loci[mut]
    return mutation_loci


def add_knockin_loci(candidate_loci: dict[str, set], knockin_loci: set):
    mutation_loci = dict()
    for mut in ["+", "-", "*"]:
        mutation_loci[mut] = candidate_loci[mut] | knockin_loci
    return mutation_loci


def transpose_mutation_loci(mutation_loci: set[int], sequence: str) -> list[set]:
    len_sequence = len(sequence)
    mutation_loci_transposed = [set() for _ in range(len_sequence)]
    for mut, idx_mutation in mutation_loci.items():
        for i, loci in enumerate(mutation_loci_transposed):
            if i in idx_mutation:
                loci.add(mut)
    return mutation_loci_transposed


def extract_mutation_loci(TEMPDIR: Path, FASTA_ALLELES: dict, SAMPLE_NAME: str, CONTROL_NAME: str, KNOCKIN_LOCI_ALLELES: dict) -> dict[str, list]:
    MUTATION_LOCI_ALLELES = dict()
    filepath_control = Path(TEMPDIR, "midsv", f"{CONTROL_NAME}_control.json")
    strand_bias = is_strand_bias(read_midsv(filepath_control))
    for allele, sequence in FASTA_ALLELES.items():
        filepath_sample = Path(TEMPDIR, "midsv", f"{SAMPLE_NAME}_{allele}.json")
        indels_sample = count_indels(read_midsv(filepath_sample), sequence)
        coverages_sample = call_coverage_on_each_base(read_midsv(filepath_sample), sequence)
        indels_sample_normalized = normalize_indels(indels_sample, coverages_sample)
        indels_kmer_sample = split_kmer(indels_sample_normalized, kmer=11)
        # Load indels_control_normalized and indels_kmer_control
        with open(Path(TEMPDIR, "mutation_loci", f"{CONTROL_NAME}_{allele}_normalized..pickle"), "rb") as f:
            indels_control_normalized = pickle.load(f)
        with open(Path(TEMPDIR, "mutation_loci", f"{CONTROL_NAME}_{allele}_kmer..pickle"), "rb") as f:
            indels_kmer_control = pickle.load(f)
        # Extract candidate mutation loci
        dissimilar_loci = extract_dissimilar_loci(indels_kmer_sample, indels_kmer_control)
        anomal_loci = extract_anomal_loci(indels_sample_normalized, indels_control_normalized)
        mutation_loci = merge_loci(dissimilar_loci, anomal_loci)
        # # Extract error loci in homopolymer regions
        # errors_in_homopolymer = extract_errors_in_homopolymer(sequence, indels_sample_normalized, indels_control_normalized, mutation_loci)
        # mutation_loci = discard_errors_in_homopolymer(mutation_loci, errors_in_homopolymer)
        if strand_bias is False:
            mutation_loci = discard_errors_on_biased_strand(read_midsv(filepath_sample), mutation_loci)
        mutation_loci = add_knockin_loci(mutation_loci, KNOCKIN_LOCI_ALLELES[allele])
        mutation_loci_transposed = transpose_mutation_loci(mutation_loci, sequence)
        MUTATION_LOCI_ALLELES[allele] = mutation_loci_transposed
    return MUTATION_LOCI_ALLELES



from __future__ import annotations

import hashlib
import re
from pathlib import Path
from urllib.error import URLError
from urllib.request import urlopen
import xml.etree.ElementTree as ET

import mappy


########################################################################
# Check if the sample is in the proper format.
########################################################################


def _validate_file_existence(input_file: str):
    if not Path(input_file).exists():
        raise FileNotFoundError(f"{input_file} is not found")


def _validate_fastq_extension(fastq_path: str):
    if not re.search(r".fastq$|.fastq.gz$|.fq$|.fq.gz$", fastq_path):
        raise ValueError(f"{fastq_path} requires extensions either 'fastq', 'fastq.gz', 'fq' or 'fq.gz'")


# Varidate if the file is in the proper format.
# See top 100 lines
def _validate_fastq_content(fastq_path: str):
    try:
        names, seqs, quals = zip(*[(n, s, q) for i, (n, s, q) in enumerate(mappy.fastx_read(fastq_path)) if i < 100])
        if not (len(names) == len(seqs) == len(quals) > 0):
            raise ValueError
    except ValueError:
        raise ValueError(f"{fastq_path} is not a FASTQ format")


def _validate_fasta_content(fasta_path: str):
    try:
        names, seqs = zip(*[(n, s) for n, s, _ in mappy.fastx_read(fasta_path)])
        if len(names) != len(seqs) or not names:
            raise ValueError
    except ValueError:
        raise ValueError(f"{fasta_path} is not a proper FASTA format")
    if len(names) != len(set(names)):
        raise ValueError(f"{fasta_path} must include unique identifiers")
    if len(seqs) != len(set(seqs)):
        raise ValueError(f"{fasta_path} must include unique DNA sequences")
    if "control" not in names:
        raise ValueError(f"One of the headers in the {fasta_path} must be '>control'")


def validate_files(SAMPLE: str, CONTROL: str, ALLELE: str) -> None:
    for file in [CONTROL, SAMPLE, ALLELE]:
        _validate_file_existence(file)
    for file in [CONTROL, SAMPLE]:
        _validate_fastq_extension(file)
        _validate_fastq_content(file)
    _validate_fasta_content(ALLELE)


########################################################################
# Check Cache
########################################################################


def exists_cached_control(control: str, tempdir: Path) -> bool:
    path_cache_hash = Path(tempdir, "cache", "control_hash.txt")
    if path_cache_hash.exists():
        current_hash = hashlib.sha256(Path(control).read_bytes()).hexdigest()
        cashed_hash = path_cache_hash.read_text()
        if current_hash == cashed_hash:
            return True
    return False


def exists_cached_genome(genome: str, tempdir: Path, exists_cache_control: bool) -> bool:
    path_cache_genome = Path(tempdir, "cache", "genome_symbol.txt")
    if genome and exists_cache_control and path_cache_genome.exists():
        cashed_genome = path_cache_genome.read_text()
        if genome == cashed_genome:
            return True
    return False


########################################################################
# Check genome and UCSC server
########################################################################


def _is_webpage_available(url: str) -> bool:
    try:
        with urlopen(url) as response:
            return 200 <= response.status < 300
    except URLError:
        return False


def get_first_available_url(urls: list[str]) -> str | None:
    return next((url for url in urls if _is_webpage_available(url)), None)


def _fetch_xml_data(url: str) -> bytes:
    """Fetch XML data from a given URL."""
    with urlopen(url) as response:
        return response.read()


def _extract_genome_ids_from_xml(xml_data: bytes) -> set:
    """Extract genome IDs from XML data."""
    root = ET.fromstring(xml_data)
    return {cc.attrib["id"] for child in root for cc in child if cc.tag == "SOURCE"}


def _get_genome_ids_in_ucsc(url_das: str) -> set:
    """Get available genome IDs in UCSC."""
    xml_data = _fetch_xml_data(url_das)
    return _extract_genome_ids_from_xml(xml_data)


def is_genome_in_ucsc_ids(genome: str, url_das: str) -> bool:
    genome_ids = _get_genome_ids_in_ucsc(url_das)
    return genome in genome_ids


def validate_genome_and_fetch_urls(genome: str) -> dict[str, str]:
    server_lists = {
        "blat": [
            "https://genome.ucsc.edu/cgi-bin/hgBlat",
            "https://genome-asia.ucsc.edu/cgi-bin/hgBlat",
            "https://genome-euro.ucsc.edu/cgi-bin/hgBlat",
        ],
        "das": [
            "https://genome.ucsc.edu/cgi-bin/das/dsn/",
            "https://genome-asia.ucsc.edu/cgi-bin/das/dsn/",
            "https://genome-euro.ucsc.edu/cgi-bin/das/dsn",
        ],
        "goldenpath": [
            "https://hgdownload.cse.ucsc.edu/goldenPath",
            "http://hgdownload-euro.soe.ucsc.edu/goldenPath",
        ],
    }

    available_servers = {key: get_first_available_url(urls) for key, urls in server_lists.items()}

    error_messages = {
        "blat": "All UCSC blat servers are currently down. Please wait for a while and try again.",
        "das": "All UCSC DAS servers are currently down. Please wait for a while and try again.",
        "goldenpath": "All UCSC GoldenPath servers are currently down. Please wait for a while and try again.",
    }

    for key, message in error_messages.items():
        if not available_servers[key]:
            raise URLError(message)

    if not is_genome_in_ucsc_ids(genome, available_servers["das"]):
        raise ValueError(f"{genome} is not listed. Available genomes are in {available_servers['das']}")

    del available_servers["das"]

    return available_servers



from DAJIN2.core.consensus.consensus import call_consensus
from DAJIN2.core.consensus.name_handler import call_allele_name
from DAJIN2.core.consensus.name_handler import update_key_by_allele_name
from DAJIN2.core.consensus.name_handler import add_key_by_allele_name
from DAJIN2.core.consensus.clust_subsetter import subset_clust



import setuptools
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

with open("requirements.txt") as requirements_file:
    install_requirements = requirements_file.read().splitlines()

setuptools.setup(
    name="DAJIN2",
    version="0.3.1",
    author="Akihiro Kuno",
    author_email="akuno@md.tsukuba.ac.jp",
    description="One-step genotyping tools for targeted long-read sequencing",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/akikuno/DAJIN2",
    install_requires=install_requirements,
    packages=setuptools.find_packages(
        where="src",
    ),
    package_dir={"": "src"},
    entry_points={"console_scripts": ["DAJIN2=DAJIN2.main:execute"]},
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
)



from __future__ import annotations

import pickle
import midsv
import random
from pathlib import Path
from itertools import groupby
from typing import Generator
from collections import defaultdict

from DAJIN2.core.clustering.make_kmer import generate_mutation_kmers
from DAJIN2.core.clustering.make_score import make_score
from DAJIN2.core.clustering.return_labels import return_labels
from DAJIN2.utils import io


def annotate_score(path_sample, mutation_score, mutation_loci, is_control=False) -> Generator[list[float]]:
    for cssplit_kmer in generate_mutation_kmers(path_sample, mutation_loci):
        score = [0 for _ in range(len(cssplit_kmer))]
        for i, (cs_kmer, mut_score) in enumerate(zip(cssplit_kmer, mutation_score)):
            if mut_score == {}:
                continue
            # Mutation sites are not considered in controls
            # because they should be sample-specific.
            if is_control and cs_kmer.split(",")[1][0] in mutation_loci[i]:
                continue
            if cs_kmer in mut_score:
                score[i] = mut_score[cs_kmer]
        yield score


def reorder_labels(labels: list[int], start: int = 0) -> list[int]:
    labels_ordered = labels.copy()
    num = start
    d = defaultdict(int)
    for i, l in enumerate(labels_ordered):
        if not d[l]:
            num += 1
            d[l] = num
        labels_ordered[i] = d[l]
    return labels_ordered


###########################################################
# main
###########################################################


def is_strand_bias(path_control) -> bool:
    count_strand = defaultdict(int)
    for m in midsv.read_jsonl(path_control):
        count_strand[m["STRAND"]] += 1
    percentage_plus = count_strand["+"] / (count_strand["+"] + count_strand["-"])
    if 0.25 < percentage_plus < 0.75:
        return False
    else:
        return True


def add_labels(classif_sample, TEMPDIR, SAMPLE_NAME, CONTROL_NAME) -> list[dict[str]]:
    labels_all = []
    max_label = 0
    strand_bias = is_strand_bias(Path(TEMPDIR, CONTROL_NAME, "midsv", "control.json"))
    classif_sample.sort(key=lambda x: x["ALLELE"])
    for allele, group in groupby(classif_sample, key=lambda x: x["ALLELE"]):
        RANDOM_NUM = random.randint(0, 10**10)
        path_knockin = Path(TEMPDIR, SAMPLE_NAME, "knockin_loci", f"{allele}.pickle")
        if path_knockin.exists():
            with open(path_knockin, "rb") as p:
                knockin_loci = pickle.load(p)
        else:
            knockin_loci = set()
        with open(Path(TEMPDIR, SAMPLE_NAME, "mutation_loci", f"{allele}.pickle"), "rb") as p:
            mutation_loci: dict[str, set[str]] = pickle.load(p)
        if all(m == set() for m in mutation_loci):
            labels = [1] * len(classif_sample)
            labels_reorder = reorder_labels(labels, start=max_label)
            max_label = max(labels_reorder)
            labels_all.extend(labels_reorder)
            continue
        path_sample = Path(TEMPDIR, SAMPLE_NAME, "clustering", f"{allele}_{RANDOM_NUM}.json")
        path_control = Path(TEMPDIR, CONTROL_NAME, "midsv", f"{allele}.json")
        io.write_jsonl(data=group, path=path_sample)
        mutation_score: list[dict[str, float]] = make_score(path_sample, path_control, mutation_loci, knockin_loci)
        scores_sample = annotate_score(path_sample, mutation_score, mutation_loci)
        scores_control = annotate_score(path_control, mutation_score, mutation_loci, is_control=True)
        path_score_sample = Path(TEMPDIR, SAMPLE_NAME, "clustering", f"{allele}_score_{RANDOM_NUM}.json")
        path_score_control = Path(TEMPDIR, CONTROL_NAME, "clustering", f"{allele}_score_{RANDOM_NUM}.json")
        io.write_jsonl(data=scores_sample, path=path_score_sample)
        io.write_jsonl(data=scores_control, path=path_score_control)
        labels = return_labels(path_score_sample, path_score_control, path_sample, strand_bias)
        labels_reorder = reorder_labels(labels, start=max_label)
        max_label = max(labels_reorder)
        labels_all.extend(labels_reorder)
        # Remove temporary files
        path_sample.unlink()
        path_score_sample.unlink()
        path_score_control.unlink()
    clust_sample = classif_sample.copy()
    for clust, label in zip(clust_sample, labels_all):
        clust["LABEL"] = label
    return clust_sample


def add_readnum(clust_sample: list[dict]) -> list[dict]:
    clust_result = clust_sample.copy()
    readnum = defaultdict(int)
    for cs in clust_result:
        readnum[cs["LABEL"]] += 1
    for cs in clust_result:
        cs["READNUM"] = readnum[cs["LABEL"]]
    return clust_result


def add_percent(clust_sample: list[dict]) -> list[dict]:
    clust_result = clust_sample.copy()
    n_sample = len(clust_result)
    percent = defaultdict(int)
    for cs in clust_result:
        percent[cs["LABEL"]] += 1 / n_sample
    percent = {key: round(val * 100, 3) for key, val in percent.items()}
    for cs in clust_result:
        cs["PERCENT"] = percent[cs["LABEL"]]
    return clust_result


def update_labels(clust_sample: list[dict]) -> list[dict]:
    """
    Allocate new labels according to the ranking by PERCENT
    """
    clust_result = clust_sample.copy()
    clust_result.sort(key=lambda x: (-x["PERCENT"], x["LABEL"]))
    new_label = 1
    prev_label = clust_result[0]["LABEL"]
    for cs in clust_result:
        if prev_label != cs["LABEL"]:
            new_label += 1
        prev_label = cs["LABEL"]
        cs["LABEL"] = new_label
    return clust_result



from __future__ import annotations

import pickle
from pathlib import Path

import midsv

from DAJIN2.core.preprocess import mapping


###########################################################
# Consider all mutations are possible in the knockin region
# 大型欠失アレルにおいては欠失配列部分がノックイン領域となるため、この領域内にある変異は全て考慮する
# TODO このknockinの考慮は必要ないかもしれない？
###########################################################


def _is_invalid_file(reference: Path, query: Path) -> bool:
    return query == reference or query.suffix != ".fasta"


def _get_knockin_loci(reference: Path, query: Path) -> set[int]:
    alignments = mapping.to_sam(reference, query, preset="splice")
    alignments = [a.split("\t") for a in alignments]
    alignments_midsv = midsv.transform(alignments, midsv=False, cssplit=True, qscore=False)[0]
    cssplits = alignments_midsv["CSSPLIT"].split(",")

    return {i for i, cs in enumerate(cssplits) if cs.startswith("-")}


def extract_knockin_loci(TEMPDIR: str | Path, SAMPLE_NAME: str) -> None:
    reference = Path(TEMPDIR, SAMPLE_NAME, "fasta", "control.fasta")
    for query in Path(TEMPDIR, SAMPLE_NAME, "fasta").iterdir():
        if _is_invalid_file(reference, query):
            continue

        allele = query.stem
        path_output = Path(TEMPDIR, SAMPLE_NAME, "knockin_loci", f"{allele}.pickle")
        if path_output.exists():
            continue

        knockin_loci = _get_knockin_loci(reference, query)

        with open(path_output, "wb") as p:
            pickle.dump(knockin_loci, p)
